/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <algorithm>
#include <cassert>
#include <cstring>
#include <disk/buffer_manager.h>
#include <exception/disk_exception.h>
#include <limits>

using namespace beedb::disk;

BufferManager::BufferManager(const std::size_t count_frames, StorageManager &space_manager,
                             std::unique_ptr<BufferReplacementStrategy> replacement_strategy)
    : _space_manager(space_manager), _replacement_strategy(std::move(replacement_strategy)), _frames(count_frames),
      _frame_information(count_frames)
{
}

BufferManager::~BufferManager()
{
    // Check no frame is pinned anymore (this would indicate programming failure).
    for (const auto &frame_information : this->_frame_information)
    {
        assert(frame_information.is_pinned() == false && "Not all pages are unpinned on shutdown.");
    }

    // Write all dirty pages back to disk.
    this->flush();
}

Page *BufferManager::pin(const Page::page_id page_id)
{
    this->_pin_sequence++;

    auto frame_information_iterator = this->frame_information(page_id);
    const bool is_frame_buffered = frame_information_iterator != this->_frame_information.end();
    if (is_frame_buffered)
    {
        // Update frame information.
        (*frame_information_iterator).increase_pin_count(this->_pin_sequence);

        const auto index = std::distance(this->_frame_information.begin(), frame_information_iterator);
        return reinterpret_cast<Page *>(&(this->_frames[index]));
    }
    else
    {
        // Find frame for the pinned page.
        const auto frame_index = this->find_frame();
        if (this->_frame_information[frame_index].is_pinned())
        {
            throw exception::EvictedPagePinnedException(frame_index);
        }

        // Write frame back, if the data was modified.
        if (this->_frame_information[frame_index].is_dirty() == true)
        {
            this->_space_manager.write(reinterpret_cast<Page &>(this->_frames[frame_index]));
        }

        // Load page into frame.
        const auto page = this->_space_manager.read(page_id);
        auto *frame = &(this->_frames[frame_index]);
        std::memcpy(frame, &page, sizeof(Page));

        // Clear frame information.
        this->_frame_information[frame_index].occupy(page.id(), this->_pin_sequence);
        this->_evicted_frames++;

        return reinterpret_cast<Page *>(frame);
    }
}

void BufferManager::unpin(const Page::page_id page_id, const bool is_dirty)
{
    auto frame_information_iterator = this->frame_information(page_id);
    if (frame_information_iterator != this->_frame_information.end())
    {
        auto &frame_information = *frame_information_iterator;
        if (frame_information.pin_count() < 1)
        {
            throw exception::PageWasNotPinnedException(page_id);
        }
        frame_information.decrease_pin_count();
        frame_information.is_dirty(frame_information.is_dirty() | is_dirty);
    }
}

Page *BufferManager::allocate()
{
    const auto page_id = this->_space_manager.allocate();
    return this->pin(page_id);
}

void BufferManager::flush()
{
    for (auto i = 0u; i < this->_frame_information.size(); i++)
    {
        auto &frame_information = this->_frame_information[i];

        // Write back, when the frame is dirty.
        if (frame_information.is_occupied() && frame_information.is_dirty())
        {
            auto &page = reinterpret_cast<Page &>(this->_frames[i]);
            assert(page.id() == frame_information.page_id());
            this->_space_manager.write(page);
            this->_frame_information[i].is_dirty(false);
        }
    }
}

std::vector<beedb::disk::FrameInformation>::iterator BufferManager::frame_information(const Page::page_id page_id)
{
    return std::find_if(this->_frame_information.begin(), this->_frame_information.end(),
                        [page_id](const FrameInformation &info) { return info.page_id() == page_id; });
}

std::size_t BufferManager::find_frame()
{
    // Check buffer for not occupied pages.
    for (auto index = 0u; index < this->_frames.size(); index++)
    {
        if (this->_frame_information[index].is_occupied() == false)
        {
            return index;
        }
    }

    // If no free buffer frame found, replace an occupied one.
    return this->_replacement_strategy->find_victim(this->_frame_information);
}