/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <expression/predicate.h>

using namespace beedb::expression;

beedb::expression::Attributes PredicateAnalyzer::attributes(const Predicate &predicate)
{
    Attributes attributes;
    if (std::holds_alternative<std::unique_ptr<LogicalConnective>>(predicate))
    {
        const auto &connective = std::get<std::unique_ptr<LogicalConnective>>(predicate);
        auto left_attributes = PredicateAnalyzer::attributes(connective->left);
        auto right_attributes = PredicateAnalyzer::attributes(connective->right);
        for (const auto &attr : left_attributes)
        {
            attributes.push_back(attr);
        }
        for (const auto &attr : right_attributes)
        {
            attributes.push_back(attr);
        }
    }
    else if (std::holds_alternative<std::unique_ptr<Atom>>(predicate))
    {
        const auto &atom = std::get<std::unique_ptr<Atom>>(predicate);
        if (std::holds_alternative<Attribute>(atom->left))
        {
            attributes.push_back(std::get<Attribute>(atom->left));
        }

        if (std::holds_alternative<Attribute>(atom->right))
        {
            attributes.push_back(std::get<Attribute>(atom->right));
        }
    }

    return attributes;
}

bool PredicateAnalyzer::contains_range_predicate(const Predicate &predicate)
{
    if (std::holds_alternative<std::unique_ptr<LogicalConnective>>(predicate))
    {
        const auto &connective = std::get<std::unique_ptr<LogicalConnective>>(predicate);
        return PredicateAnalyzer::contains_range_predicate(connective->left) ||
               PredicateAnalyzer::contains_range_predicate(connective->right);
    }
    else if (std::holds_alternative<std::unique_ptr<Atom>>(predicate))
    {
        const auto &atom = std::get<std::unique_ptr<Atom>>(predicate);
        auto atom_ptr = atom.get();
        return (typeid(*atom_ptr) == typeid(expression::EQ) || typeid(*atom_ptr) == typeid(expression::NEQ)) == false;
    }

    return false;
}

bool PredicateAnalyzer::contains_not_equals_predicate(const Predicate &predicate)
{
    if (std::holds_alternative<std::unique_ptr<LogicalConnective>>(predicate))
    {
        const auto &connective = std::get<std::unique_ptr<LogicalConnective>>(predicate);
        return PredicateAnalyzer::contains_not_equals_predicate(connective->left) ||
               PredicateAnalyzer::contains_not_equals_predicate(connective->right);
    }
    else if (std::holds_alternative<std::unique_ptr<Atom>>(predicate))
    {
        const auto &atom = std::get<std::unique_ptr<Atom>>(predicate);
        auto atom_ptr = atom.get();
        return typeid(*atom_ptr) == typeid(expression::NEQ);
    }

    return false;
}