/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include <array>
#include <config.h>
#include <cstdint>
#include <cstring>
#include <iostream>
#include <cstddef>

namespace beedb::disk
{
/**
 * Represents a page on disk, holding a header for meta information
 * and raw memory for storing data.
 * Pages can be linked logically. All linked pages contain data for the
 * same table; like a linked list of storage.
 */
class Page
{
  public:
    using page_id = std::uint64_t;
    using page_offset = std::uint16_t;
    static constexpr page_id INVALID_PAGE_ID = std::numeric_limits<Page::page_id>::max();
    static constexpr page_offset INVALID_PAGE_OFFSET = std::numeric_limits<Page::page_offset>::max();

  private:
    /**
     * Stores meta information for a page.
     */
    struct Header
    {
        Page::page_id page_id = INVALID_PAGE_ID;
        Page::page_id next_page_id = INVALID_PAGE_ID;
        Page::page_offset last_written_index = 0u;
        std::byte padding[8];
    } __attribute__((packed));

  public:
    static constexpr std::size_t data_size = Config::page_size - sizeof(Header);

    Page(const page_id id) : _data({std::byte{'\0'}})
    {
        _header.page_id = id;
    }

    ~Page() = default;

    /**
     * @return Id of this page.
     */
    inline page_id id() const
    {
        return _header.page_id;
    }

    /**
     * @return Id of the page which is logical connected to this page.
     */
    inline page_id next_page_id() const
    {
        return _header.next_page_id;
    }

    /**
     * Updates the next page id.
     * @param next_page_id Id of the next page.
     */
    inline void next_page_id(const page_id next_page_id)
    {
        _header.next_page_id = next_page_id;
    }

    /**
     * @return True, when this page has a next page.
     */
    inline bool has_next_page() const
    {
        return next_page_id() != INVALID_PAGE_ID;
    }

    /**
     * @return Number of bytes actually written to this page.
     */
    inline Page::page_offset size() const
    {
        return _header.last_written_index;
    }

    /**
     * @return Number of free bytes.
     */
    inline Page::page_offset free_space() const
    {
        return _data.size() - size();
    }

    /**
     * @return Pointer to the data stored on this page.
     */
    inline std::byte *data()
    {
        return _data.data();
    }

    /**
     * Replaces content on the page.
     *
     * @param begin First byte to write.
     * @param data  Pointer to the data to be written.
     * @param length Length of the data.
     */
    inline void overwrite(const Page::page_offset begin, const std::byte *data, const Page::page_offset length)
    {
        std::memcpy(&_data[begin], data, length);
    }

    /**
     * Adds new content at the current end of the page.
     * @param data Pointer to the data.
     * @param length Length of the data.
     * @return First byte the data was written to.
     */
    inline std::size_t append(const std::byte *data, const Page::page_offset length)
    {
        const auto begin = _header.last_written_index;
        _header.last_written_index = begin + length;
        overwrite(begin, data, length);
        return begin;
    }

    /**
     * @return True, when this is a persisted page.
     */
    operator bool() const
    {
        return _header.page_id < INVALID_PAGE_ID;
    }

    /**
     * @param index Index of data.
     * @return Pointer to the data.
     */
    std::byte *operator[](const Page::page_offset index)
    {
        return &_data[index];
    }

  private:
    Header _header;
    std::array<std::byte, Page::data_size> _data;
} __attribute__((packed));
} // namespace beedb::disk