/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include <optional>
#include <ostream>
#include <set>
#include <unordered_map>
#include <variant>
#include <vector>

namespace beedb::expression
{

using AttributeName = std::string;
using TableName = std::optional<std::string>;
using Alias = std::optional<std::string>;
using InAscendingOrder = std::optional<bool>;

enum class AttributeOrigin : int
{
    PHYSICAL = 0, // the "default"; a physical column
    MIXED = 1,    // used for Asterisk
    AGG_SUM,
    AGG_COUNT,
    AGG_AVG,
    AGG_MIN,
    AGG_MAX, // aggregations
    ARI_ADD,
    ARI_MUL,
    ARI_DIV,
    ARI_SUB // arithmetic
};

static std::string to_string(AttributeOrigin o)
{
    switch (o)
    {
    case AttributeOrigin::AGG_AVG:
        return "AVG";
    case AttributeOrigin::AGG_SUM:
        return "SUM";
    case AttributeOrigin::AGG_COUNT:
        return "COUNT";
    case AttributeOrigin::AGG_MIN:
        return "MIN";
    case AttributeOrigin::AGG_MAX:
        return "MAX";
    case AttributeOrigin::PHYSICAL:
        return "PHYSICAL";
    default:
        return "UNSUPPORTED_FUNCTION";
    }
}

// the actual type, that holds attribute information:
// <attribute name,
//  table name (optional),
//  alias (optional),
//  origin (optional),
//  order (optional)>

struct Attribute
{
    friend std::ostream &operator<<(std::ostream &stream, const Attribute &attribute);
    /// MEMBERS:
    const AttributeName name;

    const TableName table = std::nullopt;

    // Note: An alias has no meaning for optimization or consistency
    const Alias alias = {};

    // Where do values from this attribute come from, a table or an operator?
    const AttributeOrigin origin = AttributeOrigin::PHYSICAL;

    // Is this attribute (supposed to be) ordered in ascending order?
    // false => desc. order, std::nullopt => no specified order
    const InAscendingOrder order = std::nullopt;

    // string representation of an attribute. alias is prefered, if available.
    // combined_name is used for comparison (see operator implementations).
    const std::string combined_name = {alias.value_or(
        (static_cast<int>(origin) > 1) ? to_string(origin) + "(" + (table ? table.value() + "." : "") + name + ")"
                                       : (table ? table.value() + "." : "") + name)};

    /// FUNCTIONS:
    operator std::string() const
    {
        return combined_name;
    }

    // convenience constructor:
    static Attribute create(const Attribute &other, const TableName table_)
    {
        return Attribute{other.name, table_, other.alias, other.origin, other.order};
    }

    // operators to utilize stl-classes templated with Attribute

    bool operator<(const Attribute &rhs) const
    {
        // simply compare compound name
        return combined_name < rhs.combined_name;
    }

    bool operator==(const Attribute &rhs) const
    {
        // simply compare compound name. Note that an alias/tablename is
        // useful for disambiguation (e.g. in self-joins)

        return (name == rhs.name) && (table == rhs.table) && (origin == rhs.origin);
    }

    bool operator>(const Attribute &rhs) const
    {
        // simply compare compound name
        return combined_name >= rhs.combined_name;
    }
    bool operator<=(const Attribute &rhs) const
    {
        // simply compare compound name
        return combined_name < rhs.combined_name;
    }
    bool operator>=(const Attribute &rhs) const
    {
        // simply compare compound name
        return combined_name >= rhs.combined_name;
    }
    bool operator!=(const Attribute &rhs) const
    {
        return combined_name != rhs.combined_name;
    }

    bool isAsterisk() const
    {
        // gives true for "*" and "T.*"
        return name == "*" && !alias && origin == AttributeOrigin::MIXED && !order;
    }
};
/**
 * @brief The Asterisk struct is a special instance of Attribute, used for user convenience.
 * It will be resolved to a set of Attributes at logical plan creation time.
 * E.g. "SELECT * FROM ..." or "SELECT T.* FROM T, ..."
 */
struct Asterisk : Attribute
{
    Asterisk(const TableName &table_ = std::nullopt)
        : Attribute{"*", table_, std::nullopt, AttributeOrigin::MIXED, std::nullopt}
    {
    }
};

/**
 * Attributes is a set of fully quallified attributes.
 *
 * "Fully quallified" implies that each attribute also holds its source.
 * This can either be a table-name or a result of a computation
 * (i.e. aggregation). Table-origin is encoded via string, aggregation (or
 * not via type/template).
 *
 * Newly produced columns hold a generated table name for intermediate
 * tables and a generated name (such as "SUM(T.A)").
 *
 * Attributes's set-property enforces unique attributes in the collection.
 * Note that an attribute's alias is used for disambiguation.
 */
using Attributes = std::vector<Attribute>;

static inline Attributes recreateAttributes(const Attributes &other_attr, const std::string &table_name)
{
    Attributes attributes;

    for (const auto &attr : other_attr)
    {
        attributes.push_back(Attribute::create(attr, table_name));
    }

    return attributes;
}

static inline std::string to_string(const Attributes &attributes)
{
    std::string ret = "";
    for (const auto &attr : attributes)
    {
        ret += " ";
        ret += attr;
    }
    return ret;
}
} // namespace beedb::expression