/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include "expression/attribute.h"
#include "expression/predicate.h"
#include <cassert>
#include <index/type.h>
#include <sstream>
#include <table/type.h>

namespace beedb::expression
{
using namespace std::string_literals;

/**
 * A RequiredSchema is a set of required attributes. Requirements of scan operators are checked differently.
 */
using RequiredSchema = Attributes;

/**
 * This is the schema produced by an operator. It is represented as an edge "weight"/data in the graph.
 */
using OutgoingSchema = Attributes;

/**
 * This is the schema/attributes that are produced by an operator, e.g. by aggregations.
 */
using AttributeAdditions = Attributes;

/**
 * @brief The Operator struct represents operators as nodes in the graph.
 *
 * Depending on the type, operators can provide additional data, aside from
 * "required schema" and "working schema":
 *
 * Logically relevant operators:
 * CROSSPRODUCT: /
 * JOIN: predicate
 * SELECTION: predicate
 * PROJECTION: /
 * AGGREGATION: /
 * GROUPBY: /
 * QUERY: logical plan
 * ARITHMETIC: an arithmetic constant (optional)
 *
 *
 */
struct Operator
{
    Operator(const RequiredSchema requirements_, const AttributeAdditions additional_attributes_,
             const bool forwards_schema_)
        : requirements(requirements_), additional_attributes(additional_attributes_), forwards_schema(forwards_schema_)
    {
    }

    virtual ~Operator() = default;

    RequiredSchema requirements;              // attributes an operator requires
    AttributeAdditions additional_attributes; // the attributes an operator works on

    // This attribute indicates that an operator type forwards incoming
    // attributes or not. This is subclass-dependet information.
    // It is used when infering outgoing schemas of an operator. This
    // information can not be derived by the operator itself, because it
    // depends on it's position in the plan.
    const bool forwards_schema;

    /**
     * @brief inferPredicateRequirements is a utility function, that
     * extracts attributes from a predicate tree.
     *
     * Uses extractAttributes() as a helper function.
     *
     * Used in JOIN/SELECTION subclasses.
     *
     * @param pred the predicate, that contains attributes
     * @return the extracted attributes
     */
    static RequiredSchema inferPredicateRequirements(const Predicate &pred)
    {
        return PredicateAnalyzer::attributes(pred);
    }

    /**
     * @brief inferAggregationRequirements is a utility function, that
     * extracts required attributes. this is implemented by replacing
     * the origin (field within an attribute) with a physical origin.
     *
     * This can be interpreted as " Attribute 'SUM(A)' requires attribute
     * 'A' to be computable."
     *
     * Used in AGGREGATION/GROUPBY subclasses.
     *
     * @param agg_attributes the attributes that might contain aggregations
     * @return attributes with origin set to physical
     */
    static RequiredSchema inferStatementRequirements(const Attributes &agg_attributes)
    {
        Attributes requirements;
        // simply clone attributes without aggregation-origin (and physical origin instead):
        for (const auto &attr : agg_attributes)
        {
            requirements.push_back({attr.name, attr.table, attr.alias, AttributeOrigin::PHYSICAL});
        }
        return requirements;
    }

    /**
     * @brief operator std::string gives a representation of the operator.
     */
    operator std::string() const
    {
        return to_string();
    }

    /**
     * @brief clone creates an new instance of an operator.
     * @return an exact copy of this
     */
    virtual std::unique_ptr<Operator> clone() const = 0;

  protected:
    virtual std::string to_string() const = 0;
};

/**
 * @brief The TABLE struct produces (physically existing) attributes of a
 * single, specific table.
 */
struct TableOperator : Operator
{
    TableOperator(const AttributeAdditions table_attributes, const std::string &table_name)
        : Operator{{/*no requirements*/}, table_attributes, false}, _table_name(table_name), _table_alias{std::nullopt}
    {
        // consistency check: we may only produce attributes from a single
        // table
        std::set<std::string> tablenames;
        for (const auto &attribute : table_attributes)
        {

            assert(attribute.origin == AttributeOrigin::PHYSICAL);
            assert(attribute.table.has_value());

            tablenames.insert(attribute.table.value());
        }

        assert(tablenames.size() == 1); // table name/alias has to be the same for all attributes of this operator

        if (table_attributes.back().table.value() != _table_name)
        {
            _table_alias = table_attributes.back().table.value();
        }
    }

    virtual ~TableOperator() = default;

    std::string to_string() const override
    {
        return "TABLE( " + _table_name + " )";
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<TableOperator>(this->additional_attributes, this->_table_name);
    }

    const std::string &physical_table_name() const
    {
        return _table_name;
    }
    const std::optional<std::string> table_alias() const
    {
        return _table_alias;
    }

  private:
    const std::string _table_name;
    std::optional<std::string> _table_alias;
};

struct IndexScanOperator : Operator
{
    IndexScanOperator(const AttributeAdditions table_attributes, const std::string &table_name_, Predicate &&predicate_,
                      expression::Attribute indexed_attribute_)
        : Operator{inferPredicateRequirements(predicate_), table_attributes, false}, table_name(table_name_),
          predicate(std::move(predicate_)), indexed_attribute(indexed_attribute_)
    {
    }

    virtual ~IndexScanOperator() = default;

    std::string to_string() const override
    {
        return "INDEX_SCAN( " + table_name + " )";
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<IndexScanOperator>(this->additional_attributes, this->table_name,
                                                   expression::clone_predicate(predicate), indexed_attribute);
    }

    const std::string table_name;
    Predicate predicate;
    expression::Attribute indexed_attribute;
};

/**
 * @brief The CROSSPRODUCT struct is just a union of incoming attributes
 * and therefore solely defined by its placement in the graph
 */
struct CrossProductOperator : public Operator
{

    CrossProductOperator() : Operator{{}, {}, true}
    {
    }

    virtual ~CrossProductOperator() = default;

    std::string to_string() const override
    {
        return "CROSSPRODUCT"s;
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<CrossProductOperator>();
    }
};

/**
 * @brief The JOIN struct is defined by it's predicate (in
 * addition to its placement)
 */
struct JoinOperator : public Operator
{
    enum Type
    {
        NestedLoopsJoin,
        HashJoin
    };

    JoinOperator(Predicate &&join_predicate, const Type type_)
        : Operator{inferPredicateRequirements(join_predicate), {/* No additions*/}, true},
          predicate(std::move(join_predicate)), type(type_)
    {
    }

    virtual ~JoinOperator() = default;

    Predicate predicate;
    const Type type;

    std::string to_string() const override
    {
        std::string pred_str;
        std::visit(expression::LogicalConnective::TermStringifier{pred_str}, predicate);
        std::string str;
        str += (type == NestedLoopsJoin) ? "NESTEDLOOPSJOIN" : "HASHJOIN";
        str += "( ";
        str += pred_str;
        str += " )";
        return str;
    }
    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<JoinOperator>(expression::clone_predicate(this->predicate), type);
    }
};
/**
 * @brief The SELECTION struct represents a simple filter operator.
 * Requirements are derived from the specified predicate.
 */
struct SelectionOperator : public Operator
{
    // we can only select from attributes that we get => requirements:
    SelectionOperator(Predicate &&selection_predicate)
        : Operator{inferPredicateRequirements(selection_predicate), {/* No additions*/}, true},
          predicate(std::move(selection_predicate))
    {
    }

    virtual ~SelectionOperator() = default;

    Predicate predicate;

    std::string to_string() const override
    {
        std::string pred_str;
        std::visit(expression::LogicalConnective::TermStringifier{pred_str}, predicate);
        return "SELECTION( " + pred_str + " )";
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<SelectionOperator>(expression::clone_predicate(this->predicate));
    }
};
/**
 * @brief The PROJECTION struct reduces the incoming attributes to the
 * specified set.
 */
struct ProjectionOperator : public Operator
{
    // we can only project what is coming in => requirements:
    ProjectionOperator(const RequiredSchema projected_attributes)
        : Operator{projected_attributes, {/* No additions */}, false}
    {
    }

    virtual ~ProjectionOperator() = default;

    std::string to_string() const override
    {
        std::string str("PROJECTION( ");
        for (const auto &attr : this->additional_attributes)
        {
            str += static_cast<std::string>(attr) + " ";
        }
        str += ")";
        return str;
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<ProjectionOperator>(this->requirements);
    }
};

/**
 * @brief The AGGREGATION struct performs an aggregation operation and
 * creates a new column.
 *
 * Also, it forwards all other incoming columns,
 * including the one that this operation is based on.
 *
 * I.e. {"A"} x {"A*4"} -> {"A","A*4"}
 */
struct AggregationOperator : public Operator
{
    AggregationOperator(const AttributeAdditions resulting_attributes)
        : Operator{inferStatementRequirements(resulting_attributes), resulting_attributes, false}
    {
    }

    virtual ~AggregationOperator() = default;

    std::string to_string() const override
    {
        return "AGGREGATION"s;
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<AggregationOperator>(this->additional_attributes);
    }
};

/**
 * @brief The GROUPBY struct is similar to aggregation
 *
 */
struct GroupByOperator : public AggregationOperator
{
    // GROUPBY is an AGGREGATION
    GroupByOperator(const AttributeAdditions resulting_attributes) : AggregationOperator{resulting_attributes}
    {
    }

    virtual ~GroupByOperator() = default;

    std::string to_string() const override
    {
        std::string str("GROUPBY( ");
        for (const auto &attr : this->additional_attributes)
        {
            str += attr;
            str += " ";
        }
        str += ")";
        return str;
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<GroupByOperator>(this->additional_attributes);
    }
};

struct OrderByOperator : public Operator
{
    OrderByOperator(const AttributeAdditions sorted_attributes)
        : Operator{sorted_attributes, {/* No additions */}, true}
    {
        for (const auto &attribute : sorted_attributes)
        {
            assert(attribute.order.has_value()); // we can only make use of attributes that have an order specified
        }
    }

    virtual ~OrderByOperator() = default;

    std::string to_string() const override
    {
        std::string str("ORDERBY( ");
        for (const auto &attr : this->additional_attributes)
        {
            str += static_cast<std::string>(attr) + "(";
            str += (attr.order.value() ? "ASC"s : "DESC"s);
            str += ") ";
        }
        str += ")";
        return str;
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<OrderByOperator>(this->requirements);
    }
};

/**
 * @brief The ARITHMETIC struct is a binary arithmetic operation between
 * values from columns or constants.
 *
 * The type of operation itself is encoded in the specified attributes!
 */
struct ArithmeticOperator : public Operator
{
    using ArithmeticConstant = std::optional<std::variant<float, long>>;
    // An arithmetic operation is defined by the
    ArithmeticOperator(const AttributeAdditions resulting_attributes, const ArithmeticConstant constant_ = {})
        : Operator{inferStatementRequirements(resulting_attributes), resulting_attributes, true}, constant(constant_)
    {
    }

    virtual ~ArithmeticOperator() = default;

    ArithmeticConstant constant;

    std::string to_string() const override
    {
        return "ARITHMETIC"s;
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<ArithmeticOperator>(this->additional_attributes);
    }
};

struct LimitOperator : public Operator
{
    LimitOperator(const std::uint64_t limit_, const std::uint64_t offset_)
        : Operator({}, {}, true), limit(limit_), offset(offset_)
    {
    }

    virtual ~LimitOperator() = default;

    const std::uint64_t limit;
    const std::uint64_t offset;

    std::string to_string() const override
    {
        return "LIMIT( " + std::to_string(offset) + "," + std::to_string(limit) + " )";
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<LimitOperator>(limit, offset);
    }
};

/**
 * @brief The QUERY struct embeds a subquery into this plan.
 */
struct QueryOperator : public Operator
{
    // QUERY is another plan, nested into the current plan.
    QueryOperator(const RequiredSchema input_schema, const AttributeAdditions output_schema
                  //              , const LogicalPlan& nested_plan
                  )
        : Operator{input_schema, output_schema, false}
    {
    }

    virtual ~QueryOperator() = default;
    //        LogicalPlan nested_plan; // TODO: move operator defintions

    std::string to_string() const override
    {
        return "SUBQUERY"s;
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<QueryOperator>(this->requirements, this->additional_attributes);
    }
};

struct CreateTableOperator : public Operator
{
    CreateTableOperator(const std::string &table_name_, std::vector<std::string> &&column_names_,
                        std::vector<table::Type> &&column_types_, std::vector<bool> &&column_is_nullables_)
        : Operator({}, {}, false), table_name(table_name_), column_names(column_names_), column_types(column_types_),
          column_is_nullables(column_is_nullables_)
    {
    }

    virtual ~CreateTableOperator() = default;

    const std::string table_name;
    const std::vector<std::string> column_names;
    const std::vector<table::Type> column_types;
    const std::vector<bool> column_is_nullables;

    std::string to_string() const override
    {
        std::stringstream stream;
        stream << "CREATE TABLE " << table_name << " (";
        for (auto i = 0u; i < column_names.size(); i++)
        {
            stream << column_names[i] << " (" << column_types[i].name() << " ";
            if (column_is_nullables[i])
            {
                stream << "NULL";
            }
            else
            {
                stream << "NOT NULL";
            }
            stream << ")";
            if (i < column_names.size() - 1)
            {
                stream << ",";
            }
        }
        stream << " )";
        return stream.str();
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<CreateTableOperator>(table_name, std::vector<std::string>(column_names),
                                                     std::vector<table::Type>(column_types),
                                                     std::vector<bool>(column_is_nullables));
    }
};

struct CreateIndexOperator : Operator
{
    CreateIndexOperator(Attribute &&indexed_attribute, const std::string &index_name_, const bool is_unique_,
                        const index::Type index_type)
        : Operator({}, {}, false), column(indexed_attribute), index_name(index_name_), is_unique(is_unique_),
          type(index_type)
    {
    }

    virtual ~CreateIndexOperator() = default;

    const Attribute column;
    const std::string index_name;
    const bool is_unique;
    const index::Type type;

    std::string to_string() const override
    {
        return "CREATE INDEX " + index_name + " ON " + column.combined_name;
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<CreateIndexOperator>(expression::Attribute{column}, index_name, is_unique, type);
    }
};

struct InsertOperator : Operator
{
    InsertOperator(const std::string &table_name_, const Attributes columns,
                   const std::vector<std::vector<std::optional<Operand>>> &value_lists_)
        : Operator({}, columns, false), table_name(table_name_), values_lists(value_lists_)
    {
    }

    ~InsertOperator() = default;

    const std::string table_name;
    const std::vector<std::vector<std::optional<Operand>>> values_lists;

    std::string to_string() const override
    {
        std::stringstream stream;
        stream << "INSERT INTO " << table_name << " (";
        for (auto i = 0u; i < this->additional_attributes.size(); i++)
        {
            if (i > 0u)
            {
                stream << ",";
            }
            stream << this->additional_attributes[i].combined_name;
        }
        stream << ") VALUES ";

        for (auto i = 0u; i < values_lists.size(); i++)
        {
            if (i > 0u)
            {
                stream << ",";
            }
            stream << "(";
            const auto &tuple = values_lists[i];
            for (auto j = 0u; j < tuple.size(); j++)
            {
                if (j > 0u)
                {
                    stream << ",";
                }
                if (tuple[j])
                {
                    std::visit([&stream](const auto &value) { stream << value; }, tuple[j].value());
                }
                else
                {
                    stream << "NULL";
                }
            }
            stream << ")";
        }

        stream << std::flush;

        return stream.str();
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<InsertOperator>(table_name, additional_attributes, values_lists);
    }
};

struct UpdateOperator : public Operator
{
    UpdateOperator(const std::vector<std::pair<Attribute, Operand>> &updates_)
        : Operator({}, {}, false), updates(updates_)
    {
    }

    virtual ~UpdateOperator() = default;

    std::vector<std::pair<Attribute, Operand>> updates;

    std::string to_string() const override
    {
        return "UPDATE";
    }

    std::unique_ptr<Operator> clone() const override
    {
        return std::make_unique<UpdateOperator>(updates);
    }
};
} // namespace beedb::expression