/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include <chrono>
#include <database.h>
#include <plan/physical/plan.h>
#include <string>

namespace beedb::io
{
/**
 * @brief The Query struct contains the query-string and some parameters
 * that influence the execution of a query, e.g. printing the logical plan
 */
struct Query
{
    enum ExplainLevel
    {
        None,
        Plan,
        Graph
    };

    const std::string query_string;
    ExplainLevel explain = None;
};

/**
 * Wrapper for time and performance statistics, measured
 * while query execution.
 */
class ExecutionResult
{
  public:
    ExecutionResult() : _is_successful(false)
    {
    }

    ExecutionResult(const std::chrono::milliseconds build_time, const std::chrono::milliseconds execution_time,
                    const std::size_t evicted_pages)
        : _is_successful(true), _build_ms(build_time), _execution_ms(execution_time), _evicted_pages(evicted_pages)
    {
    }

    ~ExecutionResult() = default;

    std::chrono::milliseconds build_time() const
    {
        return _build_ms;
    }

    std::chrono::milliseconds execution_time() const
    {
        return _execution_ms;
    }

    std::size_t evicted_pages() const
    {
        return _evicted_pages;
    }

    bool is_successful() const
    {
        return _is_successful;
    }

  private:
    const bool _is_successful = false;
    const std::chrono::milliseconds _build_ms{};
    const std::chrono::milliseconds _execution_ms{};
    const std::size_t _evicted_pages = 0u;
};

/**
 * Executes queries and query plans.
 */
class Executor
{
  public:
    Executor(Database &database) : _database(database)
    {
    }
    ~Executor() = default;

    ExecutionResult execute(const Query &query, std::function<void(const table::Schema &)> schema_callback,
                            std::function<void(const table::Tuple &)> tuple_callback);

    ExecutionResult execute(const Query &query, std::function<void(const table::Tuple &)> tuple_callback)
    {
        return execute(
            query, [](const table::Schema &) {}, tuple_callback);
    }

    ExecutionResult execute(const Query &query)
    {
        return execute(query, [](const table::Tuple &) {});
    }

    ExecutionResult execute(plan::physical::Plan &plan, std::function<void(const table::Schema &)> schema_callback,
                            std::function<void(const table::Tuple &)> tuple_callback);

    ExecutionResult execute(plan::physical::Plan &plan, std::function<void(const table::Tuple &)> tuple_callback)
    {
        return execute(
            plan, [](const table::Schema &) {}, tuple_callback);
    }

    ExecutionResult execute(plan::physical::Plan &plan)
    {
        return execute(plan, [](const table::Tuple &) {});
    }

  protected:
    Database &_database;
};
} // namespace beedb::io