/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include <assert.h>
#include <optional>
#include <string>

#include <functional>
#include <iostream>
#include <tuple>
#include <type_traits>
#include <variant>
#include <vector>

#include "expression/predicate.h"
#include <exception/exception.h>
#include <index/type.h>
#include <table/type.h>

namespace beedb::parser
{

using SelectExpression = beedb::expression::Attributes;

/**
 * @brief The FromExpression struct is used by the QueryParser class to
 * represent expressions of tables, joins (and subqueries?).
 */
using Alias = std::optional<std::string>;
using TableDescr = std::pair<std::string, Alias>;

struct FromDescr;
using JoinDescr = std::tuple<std::unique_ptr<FromDescr>, std::unique_ptr<FromDescr>, expression::Predicate>;
using CrossProductDescr = std::vector<std::unique_ptr<FromDescr>>;
//    struct SELECTDescr;
struct FromDescr
{
    std::variant<TableDescr, JoinDescr, CrossProductDescr /*,SELECTDescr**/> descr;
};

static inline void print_structure(const std::variant<TableDescr, JoinDescr, CrossProductDescr> &from,
                                   std::string pfx = "")
{
    // std::cout << pfx;
    switch (from.index())
    {
    case 0:
        std::cout << "TableDescr" << std::endl;
        break;
    case 1: {
        std::cout << "JoinDescr with ";
        const auto &pred = std::get<expression::Predicate>(std::get<JoinDescr>(from));
        std::string pred_str;
        std::visit(expression::LogicalConnective::TermStringifier{pred_str}, pred);
        std::cout << pred_str << std::endl;
        print_structure(std::get<0>(std::get<JoinDescr>(from))->descr, pfx + "\t");
        print_structure(std::get<1>(std::get<JoinDescr>(from))->descr, pfx + "\t");
    }
    break;
    case 2:
        std::cout << "CrossProductDescr" << std::endl;
        for (const auto &el : std::get<CrossProductDescr>(from))
        {
            print_structure(el->descr, pfx + "\t");
        }
    default:
        return;
    }
}

struct FromExpression
{

    std::unique_ptr<FromDescr> from_description;

    std::vector<std::reference_wrapper<expression::Predicate>> predicates() const
    {
        auto predicates = std::vector<std::reference_wrapper<expression::Predicate>>();
        std::visit(PredicateExtractor{predicates}, from_description->descr);
        return predicates;
    }

    std::vector<std::reference_wrapper<TableDescr>> tables() const
    {
        auto attributes = std::vector<std::reference_wrapper<TableDescr>>();
        std::visit(TableExtractor{attributes}, from_description->descr);
        return attributes;
    }

    struct PredicateExtractor
    {
        // helper struct for visitor pattern
        std::vector<std::reference_wrapper<expression::Predicate>> &predicates;

        // depending on the variant type,
        // c++ picks the appropiate call operator:
        void operator()(TableDescr &)
        { /* do nothing */
        }
        void operator()(JoinDescr &join_description)
        {
            predicates.push_back(std::get<2>(join_description));
            std::visit(PredicateExtractor{predicates}, std::get<0>(join_description)->descr);
            std::visit(PredicateExtractor{predicates}, std::get<1>(join_description)->descr);
        }
        void operator()(CrossProductDescr &descriptions)
        {
            for (const auto &description : descriptions)
            {
                std::visit(PredicateExtractor{predicates}, description->descr);
            }
        }
    };

    struct TableExtractor
    {
        // helper struct for visitor pattern
        std::vector<std::reference_wrapper<TableDescr>> &attributes;

        // depending on the variant type,
        // c++ picks the appropiate call operator:
        void operator()(TableDescr &table_description)
        {
            attributes.push_back(table_description);
        }
        void operator()(JoinDescr &join_description)
        {
            std::visit(TableExtractor{attributes}, std::get<0>(join_description)->descr);
            std::visit(TableExtractor{attributes}, std::get<1>(join_description)->descr);
        }
        void operator()(CrossProductDescr &cross_product_description)
        {
            for (const auto &description : cross_product_description)
            {
                std::visit(TableExtractor{attributes}, description->descr);
            }
        }
    };
};

using WhereExpression = expression::Predicate;

using GroupByExpression = expression::Attributes;

using OrderByExpression = expression::Attributes;

struct LimitExpression
{
    std::uint64_t limit;
    std::uint64_t offset;
};

struct ColumnExpression
{
    std::string column_name;
    table::Type::Id type_id;
    std::uint16_t length;
    bool is_null;
};

struct CreateTableExpression
{
    std::string table_name;
    bool if_not_exists;
    std::vector<ColumnExpression> column_expressions;
};

struct CreateIndexExpression
{
    std::string table_name;
    std::string column_name;
    std::string index_name;
    bool is_unique;
    index::Type type;
    bool if_not_exists;
};

struct InsertExpression
{
    std::string table_name;
    std::vector<std::string> column_names;
    std::vector<std::vector<std::optional<expression::Operand>>> values_rows;
};

struct UpdateExpression
{
    std::optional<FromExpression> from;
    std::vector<std::pair<expression::Attribute, expression::Operand>> updates;
    std::optional<WhereExpression> where;
};

/**
 * @brief The QueryParser class offers methods that extract information for logical plan creation.
 * This interface can be subclassed for different types of parsers.
 * Creation of the logical plan is solely based on this interface and therefore be done without knowing the
 * frontend.
 */
class QueryParser
{

  public:
    QueryParser(const std::string &query); // simply stores the query string
    virtual ~QueryParser() = default;

    // this enum provides information about the type of query
    enum class QueryType
    {
        SELECT = 0,
        UNION,
        CREATE_TABLE,
        CREATE_INDEX,
        UPDATE,
        INSERT,
        DELETE,
        DROP,       // => "type of statements that we support"
        UNSUPPORTED // => "a given query is a statement we do not support"
    };

    QueryType type() const;

    const SelectExpression extractSELECT() const
    {
        assert(_query_type == QueryType::SELECT);
        // the implementation is guarded to prevent calling extractSelect on non-selection queries.
        return extractSELECT_impl();
    }

    const std::optional<FromExpression> extractFROM() const
    {
        assert(_query_type == QueryType::SELECT);
        return extractFROM_impl();
    }

    const std::optional<WhereExpression> extractWHERE() const
    {
        assert(_query_type == QueryType::SELECT);
        return extractWHERE_impl();
    }

    const std::optional<GroupByExpression> extractGROUPBY() const
    {
        assert(_query_type == QueryType::SELECT);
        return extractGROUPBY_impl();
    }

    const std::optional<OrderByExpression> extractORDERBY() const
    {
        assert(_query_type == QueryType::SELECT);
        return extractORDERBY_impl();
    }

    const std::optional<LimitExpression> extractLIMIT() const
    {
        assert(_query_type == QueryType::SELECT);
        return extractLIMIT_impl();
    }

    const std::optional<CreateTableExpression> extractCREATE_TABLE() const
    {
        assert(_query_type == QueryType::CREATE_TABLE);
        return extractCREATE_TABLE_impl();
    }

    const std::optional<CreateIndexExpression> extractCREATE_INDEX() const
    {
        assert(_query_type == QueryType::CREATE_INDEX);
        return extractCREATE_INDEX_impl();
    }

    const std::optional<InsertExpression> extractINSERT() const
    {
        assert(_query_type == QueryType::INSERT);
        return extractINSERT_impl();
    }

    const UpdateExpression extractUPDATE() const
    {
        assert(_query_type == QueryType::UPDATE);
        return extractUPDATE_impl();
    }

  protected:
    static bool operands_should_swap(const beedb::expression::Operand &lhs, const beedb::expression::Operand &rhs);

    virtual const SelectExpression extractSELECT_impl() const = 0;
    virtual const std::optional<FromExpression> extractFROM_impl() const = 0;
    virtual const std::optional<expression::Predicate> extractWHERE_impl() const = 0;
    virtual const std::optional<GroupByExpression> extractGROUPBY_impl() const = 0;
    virtual const std::optional<OrderByExpression> extractORDERBY_impl() const = 0;
    virtual const std::optional<LimitExpression> extractLIMIT_impl() const = 0;

    virtual const std::optional<CreateTableExpression> extractCREATE_TABLE_impl() const = 0;
    virtual const std::optional<CreateIndexExpression> extractCREATE_INDEX_impl() const = 0;

    virtual const std::optional<InsertExpression> extractINSERT_impl() const = 0;

    virtual const UpdateExpression extractUPDATE_impl() const = 0;

    /// Members:
  protected:
    const std::string &_query;
    std::optional<QueryType> _query_type = std::nullopt; // supposed to be filled in the constructor of sub-classes
};

} // namespace beedb::parser