/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include "database.h"
#include "parser/query_parser.h"
#include "plan.h"

namespace beedb::plan::logical
{

/**
 *  Used as a dictionary to resolve alias-definitions in a query.
 *  Usable for both attribute- and table alias definitions.
 *
 *  key = alias
 *  value = actual name
 */
template <typename ValueType> using AliasMapping = std::unordered_map<std::string, ValueType>;

using TableAliasMapping = AliasMapping<std::pair<table::Table *, bool>>;
using AttributeToTableNameMapping = AliasMapping<std::string>;

class Builder
{
  public:
    static const Plan build(Database &database, const parser::QueryParser &parser);
    static const Plan buildSelectPlan(Database &database, const parser::QueryParser &parser);
    static const Plan buildCreateTablePlan(Database &database, const parser::QueryParser &parser);
    static const Plan buildCreateIndexPlan(Database &database, const parser::QueryParser &parser);
    static const Plan buildInsertPlan(Database &database, const parser::QueryParser &parser);
    static const Plan buildUpdatePlan(Database &database, const parser::QueryParser &parser);

  private:
    /**
     * @brief populateAttributeToTableMappingAndCheck adds a table reference for a given attribute, if this attribute
     * has no reference yet. This action is checked and only available, if attribute can be uniquely associated with a
     * table. For example, an exception is thrown, if the user does not disambiguate the tables in a self-join.
     * @param attribute_mapping This structure is modified/populated
     * @param table_references contains references to all tables that participate in this query.
     * @param attr The attribute to be checked
     */
    static void populateAttributeToTableMappingAndCheck(AttributeToTableNameMapping &attribute_mapping,
                                                        const TableAliasMapping &table_references,
                                                        const expression::Attribute &attr);
    /**
     * @brief resolveTableReferences sets missing table references of attribute operands in the given predicate. This
     * allows the user to neglect referencing to tables, when disambiguities are resolveable (via table alias in self
     * joins or single-table queries, for example).
     * @param db an already populated map with attribute-to-table references.
     * @param pred a predicate with (potentially) missing table references
     * @return the given predicate with proper table references in attribute operands.
     */
    static expression::Predicate resolveTableReferences(const AttributeToTableNameMapping &attribute_mapping,
                                                        const expression::Predicate &pred);
    /**
     * @brief resolveTableReferences looks up the table for a given attribute and creates it anew.
     * @param attribute_mapping map, that is supposed to include a table name for this attribute
     * @param attribute the table-less attribute (that has to be recreated)
     * @return an identical attribute with a valid table reference
     */
    static const expression::Attribute resolveTableReferences(const AttributeToTableNameMapping &attribute_mapping,
                                                              const expression::Attribute &attribute);

    static const NodeIDType addFromNodes(
        const Database &db, Plan &plan,
        std::variant<parser::TableDescr, parser::JoinDescr, parser::CrossProductDescr> &tables,
        const AttributeToTableNameMapping &attribute_mapping);
    static const NodeIDType addSelection(Plan &plan, const expression::Predicate &predicate,
                                         const NodeIDType &parent_id);
    static const NodeIDType addGroupBy(const AttributeToTableNameMapping &attribute_mapping,
                                       const beedb::parser::GroupByExpression &groupbyexpr, Plan &plan,
                                       const NodeIDType &parent_id);
    static const NodeIDType addOrderBy(const AttributeToTableNameMapping &attribute_mapping,
                                       const parser::OrderByExpression &orderbyexpr, Plan &plan,
                                       const NodeIDType &parent_id);
    static const NodeIDType addLimit(const parser::LimitExpression &limit_expression, Plan &plan,
                                     const NodeIDType &parent_id);

    static void addProjection(logical::Plan &plan, const std::vector<expression::Attribute> &attributes,
                              const AttributeToTableNameMapping &attribute_mapping, const NodeIDType &parent_id);
};
} // namespace beedb::plan::logical