/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include "schema.h"
#include "tuple.h"
#include <cstring>
#include <string>
#include <vector>

namespace beedb::table
{
/**
 * Stores data in memory.
 * Some times, the data can not stay in the buffer from the BufferManager
 * and has to be copied to "safe" memory that will not be replaced.
 */
class MemoryTable
{
  public:
    MemoryTable(const Schema &schema) : _schema(schema)
    {
    }

    ~MemoryTable() = default;

    /**
     * Copy data from disk (or page) to the memory.
     *
     * @param tuple Tuple to be copied to memory.
     * @return Index within the memory table.
     */
    std::size_t copy_to_memory(const Tuple &tuple)
    {
        Tuple in_memory_tuple(_schema, _schema.row_size());
        std::memcpy(in_memory_tuple.data(), tuple.data(), _schema.row_size());
        _tuples.push_back(std::move(in_memory_tuple));
        return _tuples.size() - 1;
    }

    /**
     * @return Constant set of all tuples stored in the memory table.
     */
    const std::vector<Tuple> &tuples() const
    {
        return _tuples;
    }

    /**
     * @return Set of all tuples stored in the memory table.
     */
    std::vector<Tuple> &tuples()
    {
        return _tuples;
    }

    /**
     * @return Schema of this table.
     */
    const Schema &schema() const
    {
        return _schema;
    }

    /**
     * @return True, when the table has no tuples stored.
     */
    bool empty() const
    {
        return _tuples.empty();
    }

    /**
     * @return Number of stored tuples.
     */
    std::size_t size() const
    {
        return _tuples.size();
    }

    /**
     * Grants access to a specific tuple.
     * @param index Index of the tuple.
     * @return Tuple.
     */
    const Tuple &at(const std::size_t index) const
    {
        return _tuples[index];
    }

    const Tuple &operator[](const std::size_t index) const
    {
        return _tuples[index];
    }

  private:
    const Schema &_schema;
    std::vector<Tuple> _tuples;
};
} // namespace beedb::table