/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "column.h"
#include <vector>

#include <assert.h>

#include "expression/attribute.h"

namespace beedb::table
{

class Schema
{
  public:
    using ColumnIndexType = std::size_t;

    Schema() = default;
    Schema(const Schema &other) = default;
    Schema(Schema &&other) = default;

    /**
     * Schema creates an empty schema
     *
     * @param table_name Name of the table.
     */
    Schema(const std::string &table_name) : _table_name(table_name)
    {
    }

    /**
     * Combines to schemas to a new one.
     *
     * @param first First schema.
     * @param second Second schema.
     * @param table_name New name for the new schema (will be moved).
     */
    Schema(const Schema &first, const Schema &second, const std::string &&table_name)
        : _table_name(std::move(table_name))
    {
        const auto size = first.size() + second.size();
        _columns.reserve(size);
        _attributes.reserve(size);
        _column_order.reserve(size);
        _offset.reserve(size);

        _offset.insert(_offset.begin(), first._offset.begin(), first._offset.end());
        for (const auto offset : second._offset)
        {
            _offset.push_back(first._row_size + offset);
        }

        for (auto i = 0u; i < first.size(); i++)
        {
            _columns.push_back(first._columns[i]);
            _attributes.push_back(first._attributes[i]);
            _column_order.push_back(i);
        }

        for (auto i = 0u; i < second.size(); i++)
        {
            _columns.push_back(second._columns[i]);
            _attributes.push_back(second._attributes[i]);
            _column_order.push_back(_columns.size() - 1);
        }

        _row_size = first._row_size + second._row_size;
    }

    /**
     * @brief Schema Recreates a schema by replacing the table name with a new one. Also replaces the table-name in all
     * attributes. Warning! This is intended to be used on TABLE creation only, since attributes are only checked by
     * "name", not combined-name!
     * @param other
     * @param attributes
     * @param new_table_name the new name of the table
     */
    Schema(const Schema &other, const std::vector<expression::Attribute> &attributes, const std::string &new_table_name)
        : _table_name(new_table_name), _columns(other._columns), _attributes(attributes), _offset(other._offset),
          _column_order(other._column_order), _row_size(other._row_size)
    {

        assert(other._attributes.size() == attributes.size());
        for (auto i = 0u; i < attributes.size(); i++)
        {
            assert(attributes[i].name ==
                   other._attributes[i]
                       .name); // every attribute in "attributes" should also be included in the given schema
        }
    }

    /**
     * Creates a new schema from another schema and new attributes.
     *
     * Note that this can have a smaller length and different order then other._attributes and this->_columns!
     *
     * @param other Other schema.
     * @param attributes New attributes.
     */
    Schema(const Schema &other, const std::vector<expression::Attribute> &attributes)
        : _table_name(other._table_name), _columns(other._columns), _attributes(attributes), _offset(other._offset),
          _row_size(other._row_size)
    {
        // build _column_order based on "attributes", since it's order does not necessarily coincide with the physical
        // column order
        for (const auto &attribute : attributes)
        {

            auto old_index = other.column_index(attribute); // the "other" schema has the correct mapping!
            assert(
                old_index.has_value()); // every attribute in "attributes" should also be included in the given schema

            const auto &old_attribute = other.attribute(old_index.value());
            assert(old_attribute.table == attribute.table);

            _column_order.push_back(old_index.value());
        }
    }

    ~Schema() = default;

    /**
     * Adds a new column and its attribute to the schema.
     *
     * @param column Column to be added.
     * @param attribute Logical attribute for the column.
     * @param visible True, when the column is visible for output.
     */
    void add(Column column, expression::Attribute &&attribute)
    {
        assert(attribute.table.has_value() && attribute.table.value() == _table_name);
        _attributes.push_back(std::move(attribute));
        _columns.push_back(column);
        _column_order.push_back(_columns.size() - 1);
        if (_offset.empty())
        {
            _offset.push_back(0u);
        }
        else
        {
            const auto last_index = _offset.size() - 1;
            _offset.push_back(_offset[last_index] + _columns[last_index].type().size());
        }
        _row_size += column.type().size();
    }

    /**
     * @return Number of columns in the schema.
     */
    inline std::size_t size() const
    {
        return _columns.size();
    }

    /**
     * Calculates the byte-offset for a specific column.
     *
     * @param column_index Index of the column.
     * @return Offset in number of bytes for the raw data access.
     */
    inline std::size_t offset(const std::size_t column_index) const
    {
        return _offset[column_index];
    }

    /**
     * @return Number of bytes of the raw data.
     */
    inline std::uint32_t row_size() const
    {
        return _row_size;
    }

    /**
     *
     * Calculates the index of a column in the schema. Based on expression::Attribute equality semantics.
     * @param attribute Logical attribute.
     * @return Index in the schema.
     */
    std::optional<ColumnIndexType> column_index(const expression::Attribute &attribute) const
    {
        for (auto i = 0u; i < _attributes.size(); i++)
        {
            if (_attributes[i] == attribute)
            { // uses combined_name/expression::Attribute equality semantics
                return _column_order[i];
            }
        }
        return std::nullopt;
    }

    /**
     * Calculates the index of a column in the schema. Search is solely based on the attributes NAME,
     * neither on its table name or alias!!
     *
     * @param attribute_name Name of the column.
     * @return Index in the schema.
     */
    std::optional<ColumnIndexType> column_index(const std::string &attribute_name) const
    {
        // this version is solely based on the attribute name
        for (auto i = 0u; i < _attributes.size(); i++)
        {
            if (_attributes[i].name == attribute_name)
            {
                return _column_order[i];
            }
        }
        return std::nullopt;
    }

    /**
     * Checks whether the schema holds a specific attribute.
     *
     * @param attribute Logical attribute.
     * @return True, when the schema contains the attribute.
     */
    bool contains(const expression::Attribute &attribute) const
    {
        return column_index(attribute).has_value();
    }

    /**
     * Checks whether the schema holds a specific column.
     *
     * @param column_name Name of the column.
     * @return True, when the schema contains a column with the give name.
     */
    bool contains(const std::string &column_name) const
    {
        return column_index(column_name).has_value();
    }

    /**
     * @return Name of the table represented by this schema.
     */
    inline const std::string &table_name() const
    {
        return _table_name;
    }

    /**
     * @return Constant set of all columns.
     */
    inline const std::vector<Column> &columns() const
    {
        return _columns;
    }

    /**
     * Access to a specific column.
     *
     * @param index Index of the column.
     * @return Constant access to the column.
     */
    inline const Column &column(const std::size_t index) const
    {
        return _columns[index];
    }

    /**
     * @return Set of all attributes.
     */
    inline const std::vector<expression::Attribute> &attributes() const
    {
        return _attributes;
    }

    /**
     * Access to a specific attribute.
     *
     * @param index Index of the attribute.
     * @return Constant access to the attribute.
     */
    inline const expression::Attribute &attribute(const std::size_t index) const
    {
        return _attributes[index];
    }

    /**
     * @return True, when no column was added to the schema.
     */
    inline bool empty() const
    {
        return _columns.empty();
    }

    /**
     * @return Order of the column indices.
     */
    const std::vector<std::uint16_t> &column_order() const
    {
        return _column_order;
    }

    //        /**
    //         * Updates the column order.
    //         *
    //         * @param column_order Ordered column indices.
    //         */
    //        void column_order(std::vector<std::uint16_t>&& column_order) { _column_order = column_order; }

    Column &operator[](const ColumnIndexType index)
    {
        return _columns[index];
    }
    const Column &operator[](const ColumnIndexType index) const
    {
        return _columns[index];
    }

  private:
    const std::string _table_name;
    std::vector<Column> _columns;
    std::vector<expression::Attribute> _attributes;
    std::vector<std::uint16_t> _offset;
    std::vector<std::uint16_t> _column_order;
    std::uint32_t _row_size = 0u;
};
} // namespace beedb::table