/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "column.h"
#include "schema.h"
#include "type.h"
#include "value.h"
#include <algorithm>
#include <bitset>
#include <cassert>
#include <cstddef>
#include <cstdint>
#include <disk/page.h>
#include <memory>
#include <string>
#include <type_traits>
#include <variant>
#include <vector>

namespace beedb::table
{
/**
 * Represents a tuple stored on the database.
 */
class Tuple
{
  public:
    /**
     * Creates a tuple with a given schema, stored on a page on the disk.
     *
     * @param schema Schema for the tuple.
     * @param page_id Id of the page the tuple is stored on.
     * @param data Raw content of the tuple.
     */
    Tuple(const Schema &schema, const disk::Page::page_id page_id, const disk::Page::page_offset in_page_offset,
          std::byte *data)
        : _schema(schema), _page_id(page_id), _in_page_offset(in_page_offset), _data(data)
    {
    }

    /**
     * Creates a tuple living in the memory. The tuple is not
     * persisted on the disk.
     *
     * @param schema Schema of the tuple.
     * @param row_size Size of the tuple in bytes.
     */
    Tuple(const Schema &schema, const std::size_t row_size)
        : _schema(schema), _page_id(disk::Page::INVALID_PAGE_ID), _in_page_offset(disk::Page::INVALID_PAGE_OFFSET),
          _data(new std::byte[row_size])
    {
        std::memset(_data, '\0', row_size);
    }

    /**
     * Moves the data from the other tuple into this. The
     * other tuple will contain no data after moving.
     *
     * @param schema Schema for the tuple.
     * @param move_from Source tuple.
     */
    Tuple(const Schema &schema, Tuple &&move_from)
        : _schema(schema), _page_id(move_from.page_id()), _in_page_offset(move_from._in_page_offset),
          _data(move_from._data)
    {
        move_from.data(nullptr);
    }

    /**
     * Moves the data from the other tuple into this. The
     * other tuple will contain no data after moving.
     *
     * @param move_from Source tuple.
     */
    Tuple(Tuple &&move_from) noexcept
        : _schema(move_from.schema()), _page_id(std::move(move_from._page_id)),
          _in_page_offset(std::move(move_from._in_page_offset)), _data(std::move(move_from._data))
    {
        move_from._data = nullptr;
    }

    /**
     * Copies a tuple into memory.
     *
     * @param copy_from Source tuple.
     */
    Tuple(const Tuple &copy_from)
        : _schema(copy_from.schema()), _page_id(disk::Page::INVALID_PAGE_ID),
          _in_page_offset(disk::Page::INVALID_PAGE_OFFSET), _data(new std::byte[copy_from._schema.row_size()])
    {
        std::memcpy(_data, copy_from._data, copy_from.schema().row_size());
    }

    /**
     * Frees the memory, if the tuple is not persisted on the disk.
     */
    ~Tuple()
    {
        if (_page_id == disk::Page::INVALID_PAGE_ID && _data != nullptr)
        {
            delete[] _data;
        }
    }

    /**
     * @return Schema of the tuple.
     */
    inline const Schema &schema() const
    {
        return _schema;
    }

    /**
     * @return Id of the page the tuple is persisted on.
     */
    inline disk::Page::page_id page_id() const
    {
        return _page_id;
    }

    /**
     * @return Offset in page or -1 if tuple is in memory.
     */
    inline disk::Page::page_offset in_page_offset() const
    {
        return _in_page_offset;
    }

    /**
     * @return Access to the raw data.
     */
    inline std::byte *data() const
    {
        return _data;
    }

    /**
     * @return True, when the tuple has data.
     */
    inline bool has_data() const
    {
        return _data != nullptr;
    }

    /**
     * Updates the raw data.
     *
     * @param data New raw data.
     */
    inline void data(std::byte *data)
    {
        _data = data;
    }

    /**
     * Extracts a value from the tuple for a specific column.
     *
     * @param index Index of the column.
     * @return Value of the tuple.
     */
    Value get(const std::size_t index) const
    {
        const auto offset = _schema.offset(index);
        const Type &type = _schema[index].type();
        if (type == Type::INT)
        {
            return {type, *reinterpret_cast<std::int32_t *>(&(_data[offset]))};
        }
        else if (type == Type::LONG)
        {
            return {type, *reinterpret_cast<std::int64_t *>(&(_data[offset]))};
        }
        else if (type == Type::DOUBLE)
        {
            return {type, *reinterpret_cast<double *>(&(_data[offset]))};
        }
        else if (type == Type::FLOAT)
        {
            return {type, *reinterpret_cast<float *>(&(_data[offset]))};
        }
        else if (type == Type::CHAR)
        {
            return {type, std::string(reinterpret_cast<char *>(&_data[offset]))};
        }

        return {type, 0};
    }

    /**
     * Updates the data of the tuple for a specific column.
     *
     * @param index Index of the column.
     * @param raw_value New data.
     */
    template <typename T> void set(const std::size_t index, T &raw_value)
    {
        const auto offset = _schema.offset(index);
        if constexpr (std::is_same<T, std::string>::value)
        {
            const auto column_size = _schema[index].type().size();
            const auto length = std::min(std::size_t(column_size), raw_value.length());
            std::memcpy(&(_data[offset]), raw_value.c_str(), length);
            if (raw_value.length() < column_size)
            {
                std::memset(&(_data[offset + raw_value.length()]), '\0', column_size - raw_value.length());
            }
        }
        else if constexpr (std::is_same<T, std::nullptr_t>::value)
        {
            std::memset(&(_data[offset]), '\0', 1);
        }
        else if constexpr (std::is_same<std::remove_const_t<T>, std::int32_t>::value ||
                           std::is_same<std::remove_const_t<T>, std::int64_t>::value ||
                           std::is_same<std::remove_const_t<T>, float>::value ||
                           std::is_same<std::remove_const_t<T>, double>::value)
        {
            assert(_schema[index].type().size() == sizeof(T) && "Column and value length are different.");
            *reinterpret_cast<T *>(&(_data[offset])) = raw_value;
        }
        else
        {
            assert(false && "Unsupported type");
        }
    }

    /**
     * Set the data from a value for a specific column.
     *
     * @param index Index of the column.
     * @param value New value.
     */
    void set(const std::size_t index, const Value &value)
    {
        if (value == Type::Id::FLOAT)
        {
            auto v = value.get<float>();
            set(index, v);
        }
        else if (value == Type::Id::DOUBLE)
        {
            auto v = value.get<double>();
            set(index, v);
        }
        else if (value == Type::Id::LONG)
        {
            auto v = value.get<std::int64_t>();
            set(index, v);
        }
        else if (value == Type::Id::INT)
        {
            auto v = value.get<std::int32_t>();
            set(index, v);
        }
        else if (value == Type::Id::CHAR)
        {
            auto v = value.get<std::string>();
            set(index, v);
        }
    }

  private:
    const Schema &_schema;
    const disk::Page::page_id _page_id;
    const disk::Page::page_offset _in_page_offset;
    std::byte *_data;
};
} // namespace beedb::table