/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once
#include <cstdint>
#include <string>

namespace beedb::table
{
/**
 * Represents a database type.
 */
class Type
{
  public:
    /**
     * Ids for types.
     */
    enum Id : std::uint16_t
    {
        INT,
        LONG,
        DOUBLE,
        FLOAT,
        CHAR,
        UNKNOWN
    };

    Type(const Id id, const std::uint16_t length = 0u) : _id(id), _length(length)
    {
    }

    Type(const Type &other) = default;

    ~Type() = default;

    /**
     * @return Length of the type if the length is dynamic (like char).
     */
    inline std::uint16_t dynamic_length() const
    {
        return _length;
    }

    operator Id() const
    {
        return _id;
    }
    bool operator==(Id id) const
    {
        return _id == id;
    }
    bool operator!=(Id id) const
    {
        return _id != id;
    }

    void operator=(const Id id)
    {
        _id = id;
    }

    Type &operator=(const Type &other) = default;

    /**
     * @return Real size in bytes of the type.
     */
    std::uint16_t size() const
    {
        switch (_id)
        {
        case INT:
            return sizeof(std::int32_t);
        case LONG:
            return sizeof(std::int64_t);
        case CHAR:
            return sizeof(std::int8_t) * dynamic_length();
        case FLOAT:
            return sizeof(float);
        case DOUBLE:
            return sizeof(double);
        default:
            return 0;
        }
    }

    /**
     * @return Name of the type.
     */
    std::string name() const
    {
        switch (_id)
        {
        case INT:
            return "INT";
        case LONG:
            return "LONG";
        case DOUBLE:
            return "DOUBLE";
        case FLOAT:
            return "FLOAT";
        case CHAR:
            return "CHAR(" + std::to_string(std::int32_t(dynamic_length())) + ")";
        default:
            return "UNKNOWN";
        }
    }

  private:
    Id _id;
    std::uint16_t _length;
} __attribute__((packed));
} // namespace beedb::table