/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "type.h"
#include <cstdint>
#include <functional>
#include <ostream>
#include <variant>

namespace beedb::table
{
/**
 * Represents a value within a tuple.
 * A value contains the raw value and a type.
 * The raw value will be interpreted as the given type.
 */
class Value
{
    friend std::ostream &operator<<(std::ostream &stream, const Value &value);

  public:
    using value_type = std::variant<std::int32_t, std::int64_t, double, float, std::string>;

    Value(const Type type, const value_type value) : _type(type), _value(value)
    {
    }

    ~Value() = default;

    /**
     * @return Raw value.
     */
    inline value_type value() const
    {
        return _value;
    }

    /**
     * @return Value interpreted as given data type.
     */
    template <typename T> T get() const
    {
        return std::get<T>(_value);
    }

    /**
     * Updates the value.
     *
     * @param value New value.
     */
    inline void value(const value_type value)
    {
        _value = value;
    }

    /**
     * @return Type of the value.
     */
    inline const Type &type() const
    {
        return _type;
    }

    bool operator==(const Type &type) const
    {
        return _type == type;
    }

    bool operator==(const Value &other) const
    {
        if (_type != other._type)
        {
            return false;
        }

        return _value == other._value;
    }

    bool operator!=(const Value &other) const
    {
        if (_type != other._type)
        {
            return true;
        }

        return _value != other._value;
    }

    bool operator<=(const Value &other) const
    {
        if (_type != other._type)
        {
            return false;
        }

        return _value <= other._value;
    }

    bool operator<(const Value &other) const
    {
        if (_type != other._type)
        {
            return false;
        }

        return _value < other._value;
    }

    bool operator>=(const Value &other) const
    {
        if (_type != other._type)
        {
            return false;
        }

        return _value >= other._value;
    }

    bool operator>(const Value &other) const
    {
        if (_type != other._type)
        {
            return false;
        }

        return _value > other._value;
    }

    bool operator==(const std::nullptr_t) const
    {
        if (_type == Type::INT)
        {
            return std::get<std::int32_t>(_value) == std::numeric_limits<std::int32_t>::min();
        }
        else if (_type == Type::LONG)
        {
            return std::get<std::int64_t>(_value) == std::numeric_limits<std::int64_t>::min();
        }
        else if (_type == Type::DOUBLE)
        {
            return std::get<double>(_value) == std::numeric_limits<double>::min();
        }
        else if (_type == Type::FLOAT)
        {
            return std::get<float>(_value) == std::numeric_limits<float>::min();
        }
        else if (_type == Type::CHAR)
        {
            return std::get<std::string>(_value).empty();
        }
        else
        {
            return true;
        }
    }

    /**
     * @return Maximal value for the stored type.
     */
    Value max() const
    {
        if (_type == Type::INT)
        {
            return {_type, std::numeric_limits<std::int32_t>::max()};
        }
        else if (_type == Type::LONG)
        {
            return {_type, std::numeric_limits<std::int64_t>::max()};
        }
        else if (_type == Type::DOUBLE)
        {
            return {_type, std::numeric_limits<double>::max()};
        }
        else if (_type == Type::FLOAT)
        {
            return {_type, std::numeric_limits<float>::max()};
        }
        else if (_type == Type::CHAR)
        {
            return {_type, ""};
        }
        else
        {
            return {_type, 0};
        }
    }

    /**
     * @return Minimal value for the stored type.
     */
    Value min() const
    {
        if (_type == Type::INT)
        {
            return {_type, std::numeric_limits<std::int32_t>::min() + 1};
        }
        else if (_type == Type::LONG)
        {
            return {_type, std::numeric_limits<std::int64_t>::min() + 1};
        }
        else if (_type == Type::DOUBLE)
        {
            return {_type, std::numeric_limits<double>::min() + 1};
        }
        else if (_type == Type::FLOAT)
        {
            return {_type, std::numeric_limits<float>::min() + 1};
        }
        else if (_type == Type::CHAR)
        {
            return {_type, ""};
        }
        else
        {
            return {_type, 0};
        }
    }

    Value &operator+=(const Value &other)
    {
        if (_type == other._type)
        {
            if (_type == Type::INT)
            {
                _value = std::get<std::int32_t>(_value) + std::get<std::int32_t>(other._value);
            }
            else if (_type == Type::LONG)
            {
                _value = std::get<std::int64_t>(_value) + std::get<std::int64_t>(other._value);
            }
            else if (_type == Type::DOUBLE)
            {
                _value = std::get<double>(_value) + std::get<double>(other._value);
            }
            else if (_type == Type::FLOAT)
            {
                _value = std::get<float>(_value) + std::get<float>(other._value);
            }
        }

        return *this;
    }

    operator std::string() const
    {
        if (*this == nullptr)
        {
            return "NULL";
        }

        if (_type == Type::INT)
        {
            return std::to_string(std::get<std::int32_t>(_value));
        }
        else if (_type == Type::LONG)
        {
            return std::to_string(std::get<std::int64_t>(_value));
        }
        else if (_type == Type::DOUBLE)
        {
            return std::to_string(std::get<double>(_value));
        }
        else if (_type == Type::FLOAT)
        {
            return std::to_string(std::get<float>(_value));
        }
        else if (_type == Type::CHAR)
        {
            return std::get<std::string>(_value);
        }

        return "";
    }

  private:
    const Type _type;
    value_type _value;
};
} // namespace beedb::table

namespace std
{
template <> struct hash<beedb::table::Value>
{
  public:
    std::size_t operator()(const beedb::table::Value &value) const
    {
        if (value == beedb::table::Type::INT)
        {
            return std::hash<std::int32_t>()(value.get<std::int32_t>());
        }
        else if (value == beedb::table::Type::LONG)
        {
            return std::hash<std::int64_t>()(value.get<std::int64_t>());
        }
        else if (value == beedb::table::Type::DOUBLE)
        {
            return std::hash<double>()(value.get<double>());
        }
        else if (value == beedb::table::Type::FLOAT)
        {
            return std::hash<float>()(value.get<float>());
        }
        else if (value == beedb::table::Type::CHAR)
        {
            return std::hash<std::string>()(value.get<std::string>());
        }
        else
        {
            return 0;
        }
    }
};
} // namespace std