/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <io/printing_executor.h>
#include <io/result_output_formatter.h>
#include <util/clock.h>

using namespace beedb::io;

void PrintingExecutor::execute(const Query &query)
{
    ResultOutputFormatter result_formatter;

    auto result = Executor::execute(
        query,
        [&result_formatter](const table::Schema &schema) {
            if (schema.empty() == false)
            {
                result_formatter.header(schema);
            }
        },
        [&result_formatter](const table::Tuple &tuple) { result_formatter.push_back(tuple); });
    if (result.is_successful() && result_formatter.empty() == false)
    {
        std::cout << result_formatter;

        if (this->_database.config()[Config::k_PrintExecutionStatistics])
        {
            util::TextTable stat_table;
            stat_table.header({"Statistic", "Value"});
            stat_table.push_back({"Execution time", std::to_string(result.execution_time().count()) + " ms"});
            stat_table.push_back({"Plan time", std::to_string(result.build_time().count()) + " ms"});
            stat_table.push_back({"Fetched rows", std::to_string(result_formatter.count())});
            stat_table.push_back({"Evicted pages", std::to_string(result.evicted_pages())});
            std::cout << stat_table;
        }
        else
        {
            std::cout << "\033[0;32m"
                      << "Fetched " << result_formatter.count() << " row" << (result_formatter.count() != 1u ? "s" : "")
                      << " in " << (result.build_time().count() + result.execution_time().count()) << "ms"
                      << " \033[0m" << std::endl;
        }
    }
}