/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include "parser/hsql_parser.h"
#include <exception/parser_exception.h>

#include <regex>

using namespace beedb::parser;

HSQLParser::HSQLParser(const std::string &query) : QueryParser(query)
{
    hsql::SQLParser::parse(_query, &_hsql_result);

    if (!(_hsql_result.isValid() && _hsql_result.size() > 0))
        throw exception::SqlException(_hsql_result.errorMsg(), static_cast<std::size_t>(_hsql_result.errorLine() + 1),
                                      static_cast<std::size_t>(_hsql_result.errorColumn()));

    // determine type of query:
    _hsql_statement = _hsql_result.getStatement(0); // When do we need more then the first statement?
    switch (_hsql_statement->type())
    {
    case hsql::StatementType::kStmtSelect:
        _query_type = QueryType::SELECT;
        break;
    case hsql::StatementType::kStmtCreate: {
        auto create_type = reinterpret_cast<const hsql::CreateStatement *>(_hsql_statement)->type;
        if (create_type == hsql::CreateType::kCreateTable)
        {
            _query_type = QueryType::CREATE_TABLE;
        }
        else if (create_type == hsql::CreateType::kCreateIndex)
        {
            _query_type = QueryType::CREATE_INDEX;
        }
        break;
    }
    case hsql::StatementType::kStmtInsert:
        _query_type = QueryType::INSERT;
        break;
    case hsql::StatementType::kStmtUpdate:
        _query_type = QueryType::UPDATE;
        break;
    default:
        _query_type = QueryType::UNSUPPORTED;
    }
}

/// Definitions of interface functions:

const SelectExpression beedb::parser::HSQLParser::extractSELECT_impl() const
{
    auto select_statement = static_cast<const hsql::SelectStatement *>(_hsql_statement);
    expression::Attributes projected_columns;
    for (const hsql::Expr *column_expression : *(select_statement->selectList))
    {
        projected_columns.push_back(build_attribute(column_expression));
    }

    return projected_columns;
}

const std::optional<FromExpression> HSQLParser::extractFROM_impl() const
{
    auto table_reference = static_cast<const hsql::SelectStatement *>(_hsql_statement)->fromTable;

    if (table_reference != nullptr)
    {
        auto from =
            extractFROM_impl(table_reference); // (potentially recursive) call of helper function to fill the fexp obj.
        FromExpression from_expression{std::move(from)};
        return {std::move(from_expression)};
    }
    else
    {
        return std::nullopt;
    }
}

const std::optional<WhereExpression> HSQLParser::extractWHERE_impl() const
{

    auto where_clause = static_cast<const hsql::SelectStatement *>(_hsql_statement)->whereClause;
    if (where_clause != nullptr)
    {
        return extract_predicate(where_clause);
    }
    else
        return std::nullopt;
}

const std::optional<GroupByExpression> HSQLParser::extractGROUPBY_impl() const
{
    auto groupby_clause = static_cast<const hsql::SelectStatement *>(_hsql_statement)->groupBy;

    if (groupby_clause != nullptr && groupby_clause->having != nullptr)
    {
        throw exception::UnsupportedStatementException("HAVING");
    }

    if (groupby_clause != nullptr && groupby_clause->columns)
    {
        expression::Attributes group_by;

        for (const auto group_by_expression : *(groupby_clause->columns))
        {
            group_by.push_back(build_attribute(group_by_expression));
            std::cout << "Group by: " << group_by.back() << std::endl;
        }

        return std::optional(std::move(group_by)); // is move appropriate?
    }

    else
        return std::nullopt;
}

const std::optional<OrderByExpression> HSQLParser::extractORDERBY_impl() const
{
    auto orderby_clause = static_cast<const hsql::SelectStatement *>(_hsql_statement)->order;
    if (orderby_clause != nullptr)
    {
        expression::Attributes order_by;

        for (const auto order_by_descr : *(orderby_clause))
        {
            assert(order_by_descr && order_by_descr->expr && order_by_descr->expr->name);
            auto expr = order_by_descr->expr;
            const std::string column_name(expr->name);

            // table reference and alias defintions are optional:
            const auto table_name = expr->table == nullptr ? std::nullopt : std::optional(expr->table);
            const auto column_alias = expr->alias == nullptr ? std::nullopt : std::optional(expr->alias);
            order_by.push_back(beedb::expression::Attribute{
                std::move(column_name), std::move(table_name), std::move(column_alias),
                expression::AttributeOrigin::PHYSICAL, order_by_descr->type == hsql::OrderType::kOrderAsc});
            //            std::cout << "Order by: " <<  order_by.back() << " ASC? " <<
            //            order_by.back().order.value_or("NO ORDER DEFINED???") << std::endl;
        }

        return std::optional(std::move(order_by)); // is move appropriate?
    }

    else
        return std::nullopt;
}

const std::optional<LimitExpression> HSQLParser::extractLIMIT_impl() const
{
    const auto *select_statement = reinterpret_cast<const hsql::SelectStatement *>(this->_hsql_statement);
    if (select_statement->limit == nullptr)
    {
        return {};
    }

    const auto limit = std::uint64_t(select_statement->limit->limit->ival);
    const auto offset =
        select_statement->limit->offset == nullptr ? 0u : std::uint64_t(select_statement->limit->offset->ival);
    return {{limit, offset}};
}

/// Definitions of helper functions:

beedb::expression::AttributeOrigin HSQLParser::determine_function_type(const std::string &function_name) const
{
    std::regex count("count", std::regex::icase);
    std::regex min("min", std::regex::icase);
    std::regex max("max", std::regex::icase);
    std::regex sum("sum", std::regex::icase);
    std::regex avg("avg", std::regex::icase);
    std::smatch match;

    if (std::regex_match(function_name, match, count))
    {
        return beedb::expression::AttributeOrigin::AGG_COUNT;
    }
    else if (std::regex_match(function_name, match, min))
    {
        return beedb::expression::AttributeOrigin::AGG_MIN;
    }
    else if (std::regex_match(function_name, match, max))
    {
        return beedb::expression::AttributeOrigin::AGG_MAX;
    }
    else if (std::regex_match(function_name, match, sum))
    {
        return beedb::expression::AttributeOrigin::AGG_SUM;
    }
    else if (std::regex_match(function_name, match, avg))
    {
        return beedb::expression::AttributeOrigin::AGG_AVG;
    }
    throw exception::UnsupportedStatementException("Function not supported: \'" + function_name + "\'");
}

beedb::expression::Attribute HSQLParser::build_attribute(const hsql::Expr *h_expr) const
{

    if (h_expr->type == hsql::kExprColumnRef)
    {

        if (h_expr->name == nullptr)
        {
            throw exception::UnsupportedStatementException("Column without name");
        }
        const std::string column_name(h_expr->name);

        // table reference and alias defintions are optional:
        const auto table_name = h_expr->table == nullptr ? std::nullopt : std::optional(h_expr->table);
        const auto column_alias = h_expr->alias == nullptr ? std::nullopt : std::optional(h_expr->alias);
        return expression::Attribute{std::move(column_name), std::move(table_name), std::move(column_alias),
                                     expression::AttributeOrigin::PHYSICAL};
    }

    else if (h_expr->type == hsql::kExprStar)
    {
        const auto table_name = h_expr->table == nullptr ? std::nullopt : std::optional(h_expr->table);
        // we resolve the star expression later (when database information is available).
        // Otherwise, we have to introduce a 'bool is_star' member to expression::Attribute (or similar).
        return expression::Asterisk{std::move(table_name)};
    }

    else if (h_expr->type == hsql::kExprFunctionRef)
    {

        const auto fkt = determine_function_type(h_expr->name);

        // functions/aggregations can be renamed:
        const auto column_alias = h_expr->alias == nullptr ? std::nullopt : std::optional(h_expr->name);

        // we expect the "nested expression" to give us the column name and (potentially) a table name:
        const auto inner_attr = build_attribute(h_expr->exprList->at(0));
        return expression::Attribute{std::move(inner_attr.name), std::move(inner_attr.table), std::move(column_alias),
                                     fkt};
    }

    else if (h_expr->type == hsql::kExprLiteralString || h_expr->type == hsql::kExprLiteralInt ||
             h_expr->type == hsql::kExprLiteralNull || h_expr->type == hsql::kExprLiteralFloat)
    {
        throw exception::UnsupportedStatementException("Currently no support for literals in SELECT part (TODO)!");
    }

    throw exception::UnsupportedStatementException("Unsupported expression of type " +
                                                   std::to_string(std::int32_t(h_expr->type)));
}

std::unique_ptr<FromDescr> HSQLParser::extractFROM_impl(const hsql::TableRef *table_ref) const
{

    if (table_ref->type == hsql::kTableName)
    {
        TableDescr td{table_ref->name,
                      table_ref->alias == nullptr ? std::nullopt : std::optional(table_ref->alias->name)};
        FromDescr from{std::move(td)};
        return std::make_unique<FromDescr>(std::move(from));
    }
    else if (table_ref->type == hsql::kTableJoin)
    {

        if (table_ref->join->type != hsql::JoinType::kJoinInner)
            throw exception::UnsupportedStatementException("Only INNER joins are currently supported, sorry!");

        // recursive call on elements of the join:
        auto left_relation = extractFROM_impl(table_ref->join->left);
        auto right_relation = extractFROM_impl(table_ref->join->right);

        JoinDescr join_descriptor{std::move(left_relation), std::move(right_relation),
                                  extract_predicate(table_ref->join->condition)};
        FromDescr from{std::move(join_descriptor)};
        return std::make_unique<FromDescr>(std::move(from));
    }
    else if (table_ref->type == hsql::kTableCrossProduct)
    {
        const auto tables = table_ref->list;
        CrossProductDescr cpd;

        for (const auto table : *tables)
        {
            cpd.push_back(extractFROM_impl(table)); // recursive call on elements of the cross product
        }

        FromDescr from{std::move(cpd)};
        return std::make_unique<FromDescr>(std::move(from));
    }

    else
    {
        throw exception::UnsupportedStatementException("No support for nested SELECT's in FROM (yet), sorry!");
    }
}

beedb::expression::Predicate HSQLParser::extract_predicate(const hsql::Expr *expression) const
{

    if (expression == nullptr)
    {
        throw exception::CanNotConvertNullptrException();
    }

    if (expression->isLiteral() && expression->isBoolLiteral)
    { // check for simple truth value
        return static_cast<bool>(expression->ival);
    }
    else
    {

        // depending on the operation, we create the expression differently

        // some operations are not supported:

        switch (expression->opType)
        {
        // supported operations:
        case hsql::OperatorType::kOpOr:
        case hsql::OperatorType::kOpAnd:
        case hsql::OperatorType::kOpEquals:
        case hsql::OperatorType::kOpNotEquals:
        case hsql::OperatorType::kOpLess:
        case hsql::OperatorType::kOpLessEq:
        case hsql::OperatorType::kOpGreater:
        case hsql::OperatorType::kOpGreaterEq:
            break;
        // unsupported operations:
        default:
            throw exception::UnsupportedOperatorException(std::to_string(std::int32_t(expression->opType)));
        }
        // operation can be a logical connective...?
        switch (expression->opType)
        {
        case hsql::OperatorType::kOpOr:
            return std::make_unique<expression::OR>(HSQLParser::extract_predicate(expression->expr),
                                                    HSQLParser::extract_predicate(expression->expr2));
        case hsql::OperatorType::kOpAnd:
            return std::make_unique<expression::AND>(HSQLParser::extract_predicate(expression->expr),
                                                     HSQLParser::extract_predicate(expression->expr2));
        default:
            break;
        }

        // operation is a simple atom. We expect simple operands on both sides:
        auto left = build_operand(expression->expr);

        auto right = build_operand(expression->expr2);

        // TODO: move swap step into another layer, this is a logical operation and has nothing to do with the HSQL
        // Parser
        bool should_swap =
            operands_should_swap(left, right); // this will swap operands, if lhs is a literal and rhs is an attribute

        switch (expression->opType)
        {
        // operation is an atom:
        case hsql::OperatorType::kOpEquals:
            if (should_swap)
                return std::make_unique<expression::EQ>(right, left);
            return std::make_unique<expression::EQ>(left, right);
        case hsql::OperatorType::kOpNotEquals:
            if (should_swap)
                return std::make_unique<expression::NEQ>(right, left);
            return std::make_unique<expression::NEQ>(left, right);
        case hsql::OperatorType::kOpLess:
            if (should_swap)
                return std::make_unique<expression::GT>(right, left);
            return std::make_unique<expression::LT>(left, right);
        case hsql::OperatorType::kOpLessEq:
            if (should_swap)
                return std::make_unique<expression::GE>(right, left);
            return std::make_unique<expression::LE>(left, right);
        case hsql::OperatorType::kOpGreater:
            if (should_swap)
                return std::make_unique<expression::LT>(right, left);
            return std::make_unique<expression::GT>(left, right);
        case hsql::OperatorType::kOpGreaterEq:
            if (should_swap)
                return std::make_unique<expression::LE>(right, left);
            return std::make_unique<expression::GE>(left, right);
        default:
            break; // should never happen, all cases covered?
        }
    }
    throw exception::UnsupportedOperatorException(std::to_string(std::int32_t(expression->opType)));
}

const beedb::expression::Operand HSQLParser::build_operand(const hsql::Expr *expression) const
{
    switch (expression->type)
    {
    case hsql::kExprColumnRef: {
        auto attribute = expression::Attribute{
            expression->name, expression->table != nullptr ? std::optional(expression->table) : std::nullopt};
        return beedb::expression::Operand{attribute};
    }
    case hsql::kExprLiteralInt:
        return std::int64_t(expression->ival);
    case hsql::kExprLiteralFloat:
        return float(expression->fval);
    case hsql::kExprLiteralString:
        return std::string(expression->name);
    default:
        break;
    }
    return {};
}

const std::optional<CreateTableExpression> HSQLParser::extractCREATE_TABLE_impl() const
{
    const auto *create_statement = reinterpret_cast<const hsql::CreateStatement *>(this->_hsql_statement);
    CreateTableExpression expression{};
    expression.table_name = create_statement->tableName;
    expression.if_not_exists = create_statement->ifNotExists;
    for (auto *column : *(create_statement->columns))
    {
        ColumnExpression column_expression{};
        column_expression.column_name = column->name;
        column_expression.is_null = column->nullable;
        column_expression.length = column->type.length;

        switch (column->type.data_type)
        {
        case hsql::DataType::INT:
            column_expression.type_id = table::Type::INT;
            break;
        case hsql::DataType::LONG:
            column_expression.type_id = table::Type::LONG;
            break;
        case hsql::DataType::FLOAT:
            column_expression.type_id = table::Type::FLOAT;
            break;
        case hsql::DataType::DOUBLE:
            column_expression.type_id = table::Type::DOUBLE;
            break;
        case hsql::DataType::CHAR:
            column_expression.type_id = table::Type::CHAR;
            break;
        default:
            throw exception::UnsupportedColumnType();
        }
        expression.column_expressions.push_back(std::move(column_expression));
    }

    return {std::move(expression)};
}

const std::optional<CreateIndexExpression> HSQLParser::extractCREATE_INDEX_impl() const
{
    const auto *create_statement = reinterpret_cast<const hsql::CreateStatement *>(this->_hsql_statement);
    CreateIndexExpression expression{};
    expression.table_name = create_statement->tableName;
    expression.column_name = create_statement->viewColumns->at(0);
    expression.index_name = create_statement->indexName;
    expression.type = index::Type::BTree;
    expression.is_unique = create_statement->isUnique;
    expression.if_not_exists = create_statement->ifNotExists;

    return {std::move(expression)};
}

const std::optional<InsertExpression> HSQLParser::extractINSERT_impl() const
{
    const auto *insert_statement = reinterpret_cast<const hsql::InsertStatement *>(this->_hsql_statement);
    InsertExpression insert_expression{};
    insert_expression.table_name = insert_statement->tableName;
    for (auto column_name : *insert_statement->columns)
    {
        insert_expression.column_names.push_back(std::string(column_name));
    }

    auto value_index = 0u;
    while (value_index < insert_statement->values->size())
    {
        std::vector<std::optional<expression::Operand>> value_list;
        for (auto i = 0u; i < insert_expression.column_names.size(); i++)
        {
            const auto value = insert_statement->values->at(value_index++);
            if (value->type == hsql::kExprLiteralFloat)
            {
                value_list.push_back({{float(value->fval)}});
            }
            else if (value->type == hsql::kExprLiteralInt)
            {
                value_list.push_back({{std::int64_t(value->ival)}});
            }
            else if (value->type == hsql::kExprLiteralString)
            {
                value_list.push_back({{std::string(value->name)}});
            }
            else if (value->type == hsql::kExprLiteralNull)
            {
                value_list.push_back({});
            }
        }
        insert_expression.values_rows.push_back(std::move(value_list));
    }

    return {std::move(insert_expression)};
}

const UpdateExpression HSQLParser::extractUPDATE_impl() const
{
    const auto *update_statement = reinterpret_cast<const hsql::UpdateStatement *>(this->_hsql_statement);

    auto from = this->extractFROM_impl(update_statement->table);
    std::optional<WhereExpression> where = std::nullopt;
    if (update_statement->where != nullptr)
    {
        where = extract_predicate(update_statement->where);
    }

    std::vector<std::pair<expression::Attribute, expression::Operand>> updates;
    for (auto update_clause : *update_statement->updates)
    {
        auto value = this->build_operand(update_clause->value);
        if (std::holds_alternative<expression::Attribute>(value))
        {
            throw exception::LogicalException("Only fix values can used for updates at the moment.");
        }

        auto table_name = (update_statement->table != nullptr && update_statement->table->name != nullptr)
                              ? std::make_optional(std::string(update_statement->table->name))
                              : std::nullopt;
        auto attr = expression::Attribute{update_clause->column, table_name};
        updates.push_back(std::make_pair(attr, value));
    }

    return UpdateExpression{FromExpression{std::move(from)}, std::move(updates), std::move(where)};
}