/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <plan/optimizer/optimizer.h>

using namespace beedb::plan::logical;

Optimizer::Optimizer(const beedb::Database &database, const beedb::plan::logical::Plan &unoptimized_plan)
    : _database(database), _config(database.config()), _plan(unoptimized_plan),
      _plan_root(unoptimized_plan.find_root().value_or("ERROR_ROOT_NOT_FOUND"))
{
    assert(_plan_root != "ERROR_ROOT_NOT_FOUND");
    if (!_config[Config::k_OptimizationDisableOptimization])
    // if option is zero, nothing will happen in this class. TODO: We should use optional<Optimizer> instead of empty
    // classes..
    {
        //// Fill helper constructs for convenient access of plan components:
        fill_access_maps(); // currently unused
    }
}

const Plan Optimizer::optimize()
{
    if (!_config[Config::k_OptimizationDisableOptimization])
    // if option is zero, nothing will happen in this class. TODO: We should use optional<Optimizer> instead of empty
    // classes..
    {
        logical::Plan optimized_plan;

        /// run optimizations:

        logical::Plan::NodeRecreationFunction generate_node = [&](const NodeIDType &node) {
            const expression::Operator &plan_operator = *(_plan[node]);
            const auto &operator_type_id = typeid(plan_operator);

            auto new_node = optimized_plan.get_new_operator_ID(node /*append old node id*/);

            if (operator_type_id == typeid(expression::JoinOperator))
            {
                if (_config[Config::k_OptimizationEnableHashJoin])
                {
                    /**
                     * The conditions for replacing a join with a hash join are as follows:
                     * - The current join has to be a nested loops join (short: NLJ)
                     * - The NLJ`s predicate has to be an atom
                     * - the atom has to be a comparison on equality
                     */

                    const auto &other_join = reinterpret_cast<const expression::JoinOperator &>(plan_operator);

                    if (other_join.type == expression::JoinOperator::Type::NestedLoopsJoin &&
                        std::holds_alternative<std::unique_ptr<expression::Atom>>(other_join.predicate))
                    {
                        // TODO: refactor this...:
                        const auto &atom = std::get<std::unique_ptr<expression::Atom>>(other_join.predicate);
                        auto atom_ptr = atom.get();
                        if (typeid(*atom_ptr) == typeid(expression::EQ))
                        {
                            return logical::Plan::NodeMapContent{new_node,
                                                                 std::make_unique<expression::JoinOperator>(
                                                                     expression::clone_predicate(other_join.predicate),
                                                                     expression::JoinOperator::Type::HashJoin)};
                        }
                    }
                }
            }
            else if (operator_type_id == typeid(expression::TableOperator))
            {
                const auto &other_table = reinterpret_cast<const expression::TableOperator &>(plan_operator);
                if (_config[Config::k_OptimizationEnableIndexScan] && _plan_selection_predicate_map.size() > 0)
                {
                    for (const auto &[selection_node, predicate] : _plan_selection_predicate_map)
                    {

                        // TODO: refactor this....:
                        const auto predicate_attributes = expression::PredicateAnalyzer::attributes(predicate);
                        if (predicate_attributes.size() == 1 &&
                            expression::PredicateAnalyzer::contains_not_equals_predicate(predicate) == false)
                        {
                            // First, we test if the predicate touches attributes in other_table
                            if (std::find(other_table.additional_attributes.begin(),
                                          other_table.additional_attributes.end(),
                                          predicate_attributes[0]) != other_table.additional_attributes.end())
                            {
                                // if it does, it has to have an index for the attribute in question
                                const auto &column_index =
                                    _database[other_table.physical_table_name()].schema().column_index(
                                        predicate_attributes[0].name);

                                assert(column_index.has_value());

                                const auto &column =
                                    _database[other_table.physical_table_name()].schema().column(column_index.value());
                                if (column.is_indexed(
                                        expression::PredicateAnalyzer::contains_range_predicate(predicate)))
                                {
                                    return logical::Plan::NodeMapContent{
                                        new_node,
                                        std::make_unique<expression::IndexScanOperator>(
                                            other_table.additional_attributes, other_table.physical_table_name(),
                                            expression::clone_predicate(predicate), predicate_attributes[0])};
                                }
                            }
                        }
                    }
                }
            }

            // default case: no optimization is done for this node, just copy node
            return logical::Plan::NodeMapContent{new_node, plan_operator.clone()};
        };

        auto new_root = optimized_plan.descend_and_recreate(_plan, _plan_root, generate_node);

        return optimized_plan;
    }
    else
    {
        return {}; // empty plan objects will not be chosen over the "canonical" plan
    }
}

void Optimizer::optimization_force_hash_join()
{
}

void Optimizer::fill_access_maps()
{
    // TODO: this could be done in logical::plan
    auto ref_extractor = [&](const NodeIDType &node) {
        const expression::Operator &plan_operator = *(_plan[node]);
        const auto &operator_type_id = typeid(plan_operator);

        if (operator_type_id == typeid(expression::JoinOperator))
        {
            const auto &join = reinterpret_cast<const expression::JoinOperator &>(plan_operator);
            _plan_join_map.insert({node, join});
            _plan_join_predicate_map.insert({node, join.predicate});
        }
        else if (operator_type_id == typeid(expression::TableOperator))
        {
            _plan_table_map.insert({node, reinterpret_cast<const expression::TableOperator &>(plan_operator)});
        }
        else if (operator_type_id == typeid(expression::SelectionOperator))
        {
            const auto &selection = reinterpret_cast<const expression::SelectionOperator &>(plan_operator);
            _plan_selection_map.insert({node, selection});
            _plan_selection_predicate_map.insert({node, selection.predicate});
        }
    };

    _plan.descend_and_apply(_plan_root, ref_extractor);
}