/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2020 Databases and Information Systems Group TU Dortmund           *
 * Visit us at                                                                  *
 *             http://dbis.cs.tu-dortmund.de/cms/en/home/                       *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *                                                                              *
 * Authors (alphabetical list):                                                 *
 *          Maximilian Berens   <maximilian.berens@tu-dortmund.de>              *
 *          Roland Kühn         <roland.kuehn@cs.tu-dortmund.de>                *
 *          Jan Mühlig          <jan.muehlig@tu-dortmund.de>                    *
 *------------------------------------------------------------------------------*
 */

#include <exception/execution_exception.h>
#include <execution/add_to_index_operator.h>
#include <execution/aggregate_operator.h>
#include <execution/build_index_operator.h>
#include <execution/create_index_operator.h>
#include <execution/create_table_operator.h>
#include <execution/cross_product_operator.h>
#include <execution/hash_join_operator.h>
#include <execution/index_scan_operator.h>
#include <execution/insert_operator.h>
#include <execution/limit_operator.h>
#include <execution/nested_loops_join_operator.h>
#include <execution/order_operator.h>
#include <execution/projection_operator.h>
#include <execution/selection_operator.h>
#include <execution/sequential_scan_operator.h>
#include <execution/tuple_buffer_operator.h>
#include <execution/update_operator.h>
#include <plan/physical/builder.h>
#include <type_traits>

using namespace beedb::plan::physical;

Plan Builder::build(Database &database, const logical::Plan &logical_plan)
{
    auto root = logical_plan.find_root();
    assert(root.has_value());
    auto execution_operator = build_operator(database, logical_plan, root.value());
    assert(execution_operator && "Could not create physical operators.");
    return Plan(database, std::move(execution_operator));
}

std::unique_ptr<beedb::execution::OperatorInterface> Builder::build_operator(Database &database,
                                                                             const logical::Plan &logical_plan,
                                                                             const std::string &logical_node_name)
{
    if (logical_plan.is_empty())
    {
        return nullptr;
    }

    const auto *logical_operator = logical_plan[logical_node_name].get();
    if (typeid(*logical_operator) == typeid(expression::TableOperator))
    {
        const auto table_operator = reinterpret_cast<const expression::TableOperator *>(logical_operator);
        auto table = database.table(table_operator->physical_table_name());

        auto schema = table_operator->table_alias().has_value()
                          ? table::Schema{table->schema(), table_operator->additional_attributes,
                                          table_operator->table_alias().value()}
                          : table::Schema{table->schema()};

        return std::make_unique<execution::SequentialScanOperator>(database.config()[Config::k_ScanPageLimit], schema,
                                                                   database.buffer_manager(),
                                                                   database.table_disk_manager(), *table);
    }
    else if (typeid(*logical_operator) == typeid(expression::IndexScanOperator))
    {
        const auto index_scan_operator = reinterpret_cast<const expression::IndexScanOperator *>(logical_operator);
        auto table = database.table(index_scan_operator->table_name);
        auto schema = table::Schema{table->schema()};

        auto key_ranges = extract_key_ranges(index_scan_operator->predicate);
        bool needs_range = false;
        std::for_each(key_ranges.cbegin(), key_ranges.cend(),
                      [&needs_range](const execution::KeyRange &k) { needs_range |= !k.single_key(); });

        const auto indexed_column_name = index_scan_operator->indexed_attribute.name;
        const auto indexed_column_index = schema.column_index(indexed_column_name);
        assert(indexed_column_index.has_value());
        auto index = schema.column(indexed_column_index.value()).index(needs_range);

        return std::make_unique<execution::IndexScanOperator>(database.config()[Config::k_ScanPageLimit], schema,
                                                              database.buffer_manager(), database.table_disk_manager(),
                                                              key_ranges, index);
    }
    else if (typeid(*logical_operator) == typeid(expression::ProjectionOperator))
    {
        const auto projection_operator = reinterpret_cast<const expression::ProjectionOperator *>(logical_operator);
        auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        auto child = build_operator(database, logical_plan, in_nodes[0]);
        auto schema = table::Schema{child->schema(), projection_operator->requirements};
        auto physical_projection_operator = std::make_unique<execution::ProjectionOperator>(std::move(schema));
        physical_projection_operator->child(std::move(child));
        return physical_projection_operator;
    }
    else if (typeid(*logical_operator) == typeid(expression::SelectionOperator))
    {
        const auto selection_operator = reinterpret_cast<const expression::SelectionOperator *>(logical_operator);
        auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        auto child = build_operator(database, logical_plan, in_nodes[0]);
        auto predicate_matcher = build_predicate(selection_operator->predicate, child->schema());
        auto physical_selection_operator =
            std::make_unique<execution::SelectionOperator>(child->schema(), std::move(predicate_matcher));
        physical_selection_operator->child(std::move(child));
        return physical_selection_operator;
    }
    else if (typeid(*logical_operator) == typeid(expression::JoinOperator))
    {
        const auto join_operator = reinterpret_cast<const expression::JoinOperator *>(logical_operator);
        auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        auto left_child = build_operator(database, logical_plan, in_nodes[0]);
        auto right_child = build_operator(database, logical_plan, in_nodes[1]);
        auto join_schema =
            table::Schema(left_child->schema(), right_child->schema(),
                          left_child->schema().table_name() + "_JOIN_" + right_child->schema().table_name());

        std::unique_ptr<execution::BinaryOperator> physical_join_operator{};
        if (join_operator->type == expression::JoinOperator::HashJoin)
        {
            auto predicate_matcher =
                build_predicate(join_operator->predicate, left_child->schema(), right_child->schema());
            auto attribute_matcher =
                reinterpret_cast<execution::AttributeMatcher<execution::PredicateMatcherInterface::EQ> *>(
                    predicate_matcher.get());
            physical_join_operator = std::make_unique<execution::HashJoinOperator>(
                std::move(join_schema), attribute_matcher->left_index(), attribute_matcher->right_index());
        }
        else
        {
            auto predicate_matcher =
                build_predicate(join_operator->predicate, left_child->schema(), right_child->schema());
            physical_join_operator = std::make_unique<execution::NestedLoopsJoinOperator>(std::move(join_schema),
                                                                                          std::move(predicate_matcher));
        }

        physical_join_operator->left_child(std::move(left_child));
        physical_join_operator->right_child(std::move(right_child));
        return physical_join_operator;
    }
    else if (typeid(*logical_operator) == typeid(expression::CrossProductOperator))
    {
        auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        auto left_child = build_operator(database, logical_plan, in_nodes[0]);
        auto right_child = build_operator(database, logical_plan, in_nodes[1]);
        auto join_schema =
            table::Schema(left_child->schema(), right_child->schema(),
                          left_child->schema().table_name() + "_CROSS_" + right_child->schema().table_name());
        auto physical_crossproduct_operator = std::make_unique<execution::CrossProductOperator>(std::move(join_schema));
        physical_crossproduct_operator->left_child(std::move(left_child));
        physical_crossproduct_operator->right_child(std::move(right_child));
        return physical_crossproduct_operator;
    }
    else if (typeid(*logical_operator) == typeid(expression::GroupByOperator))
    {

        // TODO: implement this (by uncommenting, when operator is implemented)
        //        const auto groupby_operator = reinterpret_cast<const expression::GROUPBY*>(logical_operator);
        //        auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        //        auto child = build_operator(database,logical_plan, in_nodes[0]);
        /// WARNING, WHEN IMPLEMENTING: child->schema is not the proper schema, since this operator changes it/has
        /// different attributes
        //        auto physical_groupby_operator = std::make_unique<execution::AggregateOperator>(child->schema(),
        //        groupby_operator->attributes); physical_groupby_operator->child(std::move(child));
        ////        return physical_groupby_operator;
        throw exception::NoPhysicalOperatorForNode("GROUP BY");
    }
    else if (typeid(*logical_operator) == typeid(expression::AggregationOperator))
    {

        // TODO: implement this (by uncommenting, when operator is implemented)
        //        const auto aggregation_operator = reinterpret_cast<const expression::AGGREGATION*>(logical_operator);
        //       auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        //        auto child = build_operator(database,logical_plan, in_nodes[0]);
        //        auto physical_aggregation_operator = std::make_unique<execution::AggregateOperator>(child->schema(),
        //        expression::Attributes{}, aggregation_operator->attributes);
        //        physical_aggregation_operator->child(std::move(child));
        //        return physical_aggregation_operator;
        throw exception::NoPhysicalOperatorForNode("AGGREGATE");
    }
    else if (typeid(*logical_operator) == typeid(expression::OrderByOperator))
    {
        const auto orderby_operator = reinterpret_cast<const expression::OrderByOperator *>(logical_operator);
        auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        auto child = build_operator(database, logical_plan, in_nodes[0]);
        std::vector<std::pair<std::uint32_t, bool>> sort_indices;
        for (const auto &attribute : orderby_operator->requirements)
        {
            assert(attribute.order.has_value()); // we expect to always have an order defined here
            auto attribute_index = child->schema().column_index(attribute);
            assert(attribute_index.has_value());
            sort_indices.push_back({attribute_index.value(), attribute.order.value()});
        }
        auto physical_orderby_operator =
            std::make_unique<execution::OrderOperator>(child->schema(), std::move(sort_indices));
        physical_orderby_operator->child(std::move(child));
        return physical_orderby_operator;
    }
    else if (typeid(*logical_operator) == typeid(expression::LimitOperator))
    {
        const auto limit_operator = reinterpret_cast<const expression::LimitOperator *>(logical_operator);
        auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        auto child = build_operator(database, logical_plan, in_nodes[0]);
        auto physical_limit_operator =
            std::make_unique<execution::LimitOperator>(child->schema(), limit_operator->limit, limit_operator->offset);
        physical_limit_operator->child(std::move(child));
        return physical_limit_operator;
    }
    else if (typeid(*logical_operator) == typeid(expression::CreateTableOperator))
    {
        const auto create_table_operator = reinterpret_cast<const expression::CreateTableOperator *>(logical_operator);
        auto schema = table::Schema{create_table_operator->table_name};
        auto count_columns = create_table_operator->column_names.size();
        for (auto i = 0u; i < count_columns; i++)
        {
            schema.add({create_table_operator->column_types[i], create_table_operator->column_is_nullables[i]},
                       {create_table_operator->column_names[i], create_table_operator->table_name});
        }

        return std::make_unique<execution::CreateTableOperator>(database, std::move(schema));
    }
    else if (typeid(*logical_operator) == typeid(expression::CreateIndexOperator))
    {
        const auto create_index_operator = reinterpret_cast<const expression::CreateIndexOperator *>(logical_operator);
        auto physical_create_index_operator = std::make_unique<execution::CreateIndexOperator>(
            database, create_index_operator->column.table.value(), create_index_operator->column,
            create_index_operator->index_name, create_index_operator->is_unique, create_index_operator->type);
        auto table = database.table(create_index_operator->column.table.value());
        auto physical_scan_operator = std::make_unique<execution::SequentialScanOperator>(
            database.config()[Config::k_ScanPageLimit], table->schema(), database.buffer_manager(),
            database.table_disk_manager(), *table);
        auto column_index = physical_scan_operator->schema().column_index(create_index_operator->column.name);
        assert(column_index.has_value());

        auto physical_build_index_operator =
            std::make_unique<execution::BuildIndexOperator>(database, create_index_operator->column.table.value(),
                                                            column_index.value(), create_index_operator->index_name);
        physical_build_index_operator->create_index_operator(std::move(physical_create_index_operator));
        physical_build_index_operator->data_operator(std::move(physical_scan_operator));
        return physical_build_index_operator;
    }
    else if (typeid(*logical_operator) == typeid(expression::InsertOperator))
    {
        const auto insert_operator = reinterpret_cast<const expression::InsertOperator *>(logical_operator);
        auto table = database.table(insert_operator->table_name);
        auto tuple_buffer_operator = std::make_unique<execution::TupleBufferOperator>(table->schema());

        for (const auto &values : insert_operator->values_lists)
        {
            auto tuple = build_tuple(tuple_buffer_operator->schema(), insert_operator->additional_attributes, values);
            tuple_buffer_operator->add(tuple);
        }

        auto root_operator = std::unique_ptr<execution::UnaryOperator>(new execution::InsertOperator(
            database.buffer_manager(), database.table_disk_manager(), database.system_statistics(), *table));
        root_operator->child(std::move(tuple_buffer_operator));

        for (auto i = 0u; i < table->schema().size(); i++)
        {
            const auto &column = table->schema().column(i);
            for (auto index : column.indices())
            {
                auto add_to_index_operator =
                    std::unique_ptr<execution::UnaryOperator>(new execution::AddToIndexOperator(i, index));
                add_to_index_operator->child(std::move(root_operator));
                root_operator = std::move(add_to_index_operator);
            }
        }

        return root_operator;
    }
    else if (typeid(*logical_operator) == typeid(expression::UpdateOperator))
    {
        const auto update_operator = reinterpret_cast<const expression::UpdateOperator *>(logical_operator);
        auto in_nodes = logical_plan.incoming_nodes(logical_node_name);
        auto child = build_operator(database, logical_plan, in_nodes[0]);

        std::vector<std::pair<table::Schema::ColumnIndexType, table::Value>> new_column_values;
        for (const auto &attribute_value_pair : update_operator->updates)
        {
            const auto index = child->schema().column_index(attribute_value_pair.first);
            if (index.has_value())
            {
                new_column_values.push_back({index.value(), build_value(attribute_value_pair.second,
                                                                        child->schema().column(index.value()).type())});
            }
        }

        auto physical_update_operator =
            std::make_unique<execution::UpdateOperator>(database.table_disk_manager(), new_column_values);
        physical_update_operator->child(std::move(child));
        return physical_update_operator;
    }

    return {};
}

Plan Builder::build_index_plan(Database &database, const std::string &table_name, const std::string &column_name,
                               const std::string &index_name)
{
    auto table = database.table(table_name);
    auto physical_scan_operator = std::make_unique<execution::SequentialScanOperator>(
        database.config()[Config::k_ScanPageLimit], table->schema(), database.buffer_manager(),
        database.table_disk_manager(), *table);
    const auto column_index = physical_scan_operator->schema().column_index(column_name);
    assert(column_index.has_value());

    auto physical_build_index_operator =
        std::make_unique<execution::BuildIndexOperator>(database, table_name, column_index.value(), index_name);

    physical_build_index_operator->data_operator(std::move(physical_scan_operator));

    return Plan(database, std::move(physical_build_index_operator));
}

template <class... Ts> struct overloaded : Ts...
{
    using Ts::operator()...;
};
template <class... Ts> overloaded(Ts...)->overloaded<Ts...>;
std::unique_ptr<beedb::execution::PredicateMatcherInterface> Builder::build_predicate(
    const beedb::expression::Predicate &predicate, const table::Schema &schema)
{
    std::unique_ptr<execution::PredicateMatcherInterface> matcher{};
    std::visit(

        overloaded{[&schema, &matcher](const std::unique_ptr<expression::LogicalConnective> &connective) {
                       auto connective_ptr = connective.get();
                       if (typeid(*connective_ptr) == typeid(expression::AND))
                       {
                           matcher = std::make_unique<execution::AndMatcher>(
                               build_predicate(connective->left, schema), build_predicate(connective->right, schema));
                       }
                       if (typeid(*connective_ptr) == typeid(expression::OR))
                       {
                           matcher = std::make_unique<execution::OrMatcher>(build_predicate(connective->left, schema),
                                                                            build_predicate(connective->right, schema));
                       }
                   },
                   [&schema, &matcher](const std::unique_ptr<expression::Atom> &atom) {
                       matcher = build_predicate(atom, schema);
                   },
                   [](bool) {
                       /* Nothing to deal with */
                   }},

        predicate);

    return matcher;
}

std::unique_ptr<beedb::execution::PredicateMatcherInterface> Builder::build_predicate(
    const beedb::expression::Predicate &predicate, const beedb::table::Schema &left_schema,
    const beedb::table::Schema &right_schema)
{
    std::unique_ptr<execution::PredicateMatcherInterface> matcher{};

    std::visit(overloaded{[&left_schema, &right_schema,
                           &matcher](const std::unique_ptr<expression::LogicalConnective> &connective) {
                              auto connective_ptr = connective.get();
                              if (typeid(*connective_ptr) == typeid(expression::AND))
                              {
                                  matcher = std::make_unique<execution::AndMatcher>(
                                      build_predicate(connective->left, left_schema, right_schema),
                                      build_predicate(connective->right, left_schema, right_schema));
                              }
                              if (typeid(*connective_ptr) == typeid(expression::OR))
                              {
                                  matcher = std::make_unique<execution::OrMatcher>(
                                      build_predicate(connective->left, left_schema, right_schema),
                                      build_predicate(connective->right, left_schema, right_schema));
                              }
                          },
                          [&left_schema, &right_schema, &matcher](const std::unique_ptr<expression::Atom> &atom) {
                              matcher = build_predicate(atom, left_schema, right_schema);
                          },
                          [](bool) {
                              /* Nothing to deal with */
                          }},
               predicate);

    return matcher;
}

std::unique_ptr<beedb::execution::PredicateMatcherInterface> Builder::build_predicate(
    const std::unique_ptr<beedb::expression::Atom> &atom, const table::Schema &schema)
{
    const auto &left = atom->left;
    const auto &right = atom->right;

    const auto left_is_attribute = std::holds_alternative<expression::Attribute>(left);
    const auto right_is_attribute = std::holds_alternative<expression::Attribute>(right);

    assert(
        (left_is_attribute && !right_is_attribute) // non-attributes are always rhs
        ||
        (!left_is_attribute &&
         !right_is_attribute) // actually, this is constant evaluation... TODO: remove, when optimized in logical plan
        || (left_is_attribute && right_is_attribute));

    // type of left_index is std::optional<table::Schema::ColumnIndexType>
    std::optional<table::Schema::ColumnIndexType> left_index = std::nullopt;
    if (left_is_attribute)
    {
        left_index = schema.column_index(std::get<expression::Attribute>(left));
    }
    assert(left_is_attribute == left_index.has_value()); // if it is an attribute, we expect to find it

    std::optional<table::Schema::ColumnIndexType> right_index = std::nullopt;
    if (right_is_attribute)
    {
        right_index = schema.column_index(std::get<expression::Attribute>(right));
    }
    assert(right_is_attribute == right_index.has_value()); // if it is an attribute, we expect to find it

    auto atom_ptr = atom.get();
    if (typeid(*atom_ptr) == typeid(expression::EQ))
    {
        if (left_index.has_value() && right_index.has_value())
        {
            return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::EQ>>(
                left_index.value(), right_index.value());
        }
        else if (left_index.has_value() && right_index.has_value() == false)
        {
            return std::make_unique<execution::AttributeValueMatcher<execution::PredicateMatcherInterface::EQ>>(
                left_index.value(), build_value(right, schema[left_index.value()].type()));
        }
    }
    else if (typeid(*atom_ptr) == typeid(expression::LE))
    {
        if (left_index.has_value() && right_index.has_value())
        {
            return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::LE>>(
                left_index.value(), right_index.value());
        }
        else if (left_index.has_value() && right_index.has_value() == false)
        {
            return std::make_unique<execution::AttributeValueMatcher<execution::PredicateMatcherInterface::LE>>(
                left_index.value(), build_value(right, schema[left_index.value()].type()));
        }
    }
    else if (typeid(*atom_ptr) == typeid(expression::LT))
    {
        if (left_index.has_value() && right_index.has_value())
        {
            return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::LT>>(
                left_index.value(), right_index.value());
        }
        else if (left_index.has_value() && right_index.has_value() == false)
        {
            return std::make_unique<execution::AttributeValueMatcher<execution::PredicateMatcherInterface::LT>>(
                left_index.value(), build_value(right, schema[left_index.value()].type()));
        }
    }
    else if (typeid(*atom_ptr) == typeid(expression::GE))
    {
        if (left_index.has_value() && right_index.has_value())
        {
            return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::GE>>(
                left_index.value(), right_index.value());
        }
        else if (left_index.has_value() && right_index.has_value() == false)
        {
            return std::make_unique<execution::AttributeValueMatcher<execution::PredicateMatcherInterface::GE>>(
                left_index.value(), build_value(right, schema[left_index.value()].type()));
        }
    }
    else if (typeid(*atom_ptr) == typeid(expression::GT))
    {
        if (left_index.has_value() && right_index.has_value())
        {
            return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::GT>>(
                left_index.value(), right_index.value());
        }
        else if (left_index.has_value() && right_index.has_value() == false)
        {
            return std::make_unique<execution::AttributeValueMatcher<execution::PredicateMatcherInterface::GT>>(
                left_index.value(), build_value(right, schema[left_index.value()].type()));
        }
    }
    else
    {
        if (left_index.has_value() && right_index.has_value())
        {
            return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::NEQ>>(
                left_index.value(), right_index.value());
        }
        else if (left_index.has_value() && right_index.has_value() == false)
        {
            return std::make_unique<execution::AttributeValueMatcher<execution::PredicateMatcherInterface::NEQ>>(
                left_index.value(), build_value(right, schema[left_index.value()].type()));
        }
    }

    // Fallback (should never happen)
    return std::make_unique<execution::AlwaysTrueMatcher>();
}

std::unique_ptr<beedb::execution::PredicateMatcherInterface> Builder::build_predicate(
    const std::unique_ptr<beedb::expression::Atom> &atom, const beedb::table::Schema &left_schema,
    const beedb::table::Schema &right_schema)
{
    auto left = atom->left;
    auto right = atom->right;

    const auto is_left_attribute = std::holds_alternative<expression::Attribute>(left);
    const auto is_right_attribute = std::holds_alternative<expression::Attribute>(right);
    assert(is_left_attribute && is_right_attribute);

    auto left_index = left_schema.column_index(std::get<expression::Attribute>(left));
    auto right_index = right_schema.column_index(std::get<expression::Attribute>(right));

    if (left_index.has_value() == false && right_index.has_value() == false)
    {
        left_index = left_schema.column_index(std::get<expression::Attribute>(right));
        right_index = right_schema.column_index(std::get<expression::Attribute>(left));
    }

    assert(left_index.has_value());
    assert(right_index.has_value());

    auto atom_ptr = atom.get();
    if (typeid(*atom_ptr) == typeid(expression::EQ))
    {
        return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::EQ>>(
            left_index.value(), right_index.value());
    }
    else if (typeid(*atom_ptr) == typeid(expression::LE))
    {
        return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::LE>>(
            left_index.value(), right_index.value());
    }
    else if (typeid(*atom_ptr) == typeid(expression::LT))
    {
        return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::LT>>(
            left_index.value(), right_index.value());
    }
    else if (typeid(*atom_ptr) == typeid(expression::GE))
    {
        return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::GE>>(
            left_index.value(), right_index.value());
    }
    else if (typeid(*atom_ptr) == typeid(expression::GT))
    {
        return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::GT>>(
            left_index.value(), right_index.value());
    }
    else
    {
        return std::make_unique<execution::AttributeMatcher<execution::PredicateMatcherInterface::NEQ>>(
            left_index.value(), right_index.value());
    }
}

beedb::table::Value Builder::build_value(const beedb::expression::Operand &operand, const table::Type &type)
{
    table::Value::value_type value;
    std::visit(
        [&value, &type](auto &&op) {
            using T = std::decay_t<decltype(op)>;
            if constexpr (std::is_same<std::string, T>::value)
            {
                value = op;
            }
            else if constexpr (std::is_same<std::int64_t, T>::value)
            {
                if (type == table::Type::INT)
                {
                    value = static_cast<std::int32_t>(op);
                }
                else if (type == table::Type::LONG)
                {
                    value = static_cast<std::int64_t>(op);
                }
            }
            else if constexpr (std::is_same<float, T>::value)
            {
                if (type == table::Type::FLOAT)
                {
                    value = static_cast<float>(op);
                }
                else if (type == table::Type::DOUBLE)
                {
                    value = static_cast<double>(op);
                }
            }
        },
        operand);

    return table::Value{type, value};
}

beedb::table::Tuple Builder::build_tuple(const beedb::table::Schema &schema,
                                         const beedb::expression::Attributes &attributes,
                                         const std::vector<std::optional<expression::Operand>> &values)
{
    table::Tuple tuple(schema, schema.row_size());

    std::set<std::uint32_t> set_index;
    for (auto i = 0u; i < attributes.size(); i++)
    {
        const auto &attribute = attributes[i];
        const auto index = schema.column_index(attribute);
        assert(index.has_value());
        if (values[i].has_value() == false)
        {
            continue;
        }
        set_index.insert(index.value());
        const auto &column = schema.column(index.value());
        if (column == table::Type::INT)
        {
            auto value = std::int32_t(std::get<std::int64_t>(values[i].value()));
            tuple.set(index.value(), value);
        }
        else if (column == table::Type::LONG)
        {
            auto value = std::get<std::int64_t>(values[i].value());
            tuple.set(index.value(), value);
        }
        else if (column == table::Type::FLOAT)
        {
            auto value = std::get<float>(values[i].value());
            tuple.set(index.value(), value);
        }
        else if (column == table::Type::DOUBLE)
        {
            auto value = double(std::get<float>(values[i].value()));
            tuple.set(index.value(), value);
        }
        else if (column == table::Type::CHAR)
        {
            auto value = std::get<std::string>(values[i].value());
            tuple.set(index.value(), value);
        }
    }

    // Set values to NULL not given by the INSERT statement
    for (auto i = 0u; i < schema.size(); i++)
    {
        if (set_index.find(i) == set_index.end())
        {
            const auto &column = schema.column(i);
            if (column == table::Type::INT)
            {
                auto value = std::numeric_limits<std::int32_t>::min();
                tuple.set(i, value);
            }
            else if (column == table::Type::LONG)
            {
                auto value = std::numeric_limits<std::int64_t>::min();
                tuple.set(i, value);
            }
            else if (column == table::Type::DOUBLE)
            {
                auto value = std::numeric_limits<double>::min();
                tuple.set(i, value);
            }
            else if (column == table::Type::FLOAT)
            {
                auto value = std::numeric_limits<float>::min();
                tuple.set(i, value);
            }
            else if (column == table::Type::CHAR)
            {
                auto value = nullptr;
                tuple.set(i, value);
            }
        }
    }

    return tuple;
}

std::set<beedb::execution::KeyRange> Builder::extract_key_ranges(const beedb::expression::Predicate &predicate)
{
    std::set<execution::KeyRange> key_ranges;

    std::visit(overloaded{[&key_ranges](const std::unique_ptr<expression::LogicalConnective> &connective) {
                              auto connective_ptr = connective.get();
                              if (typeid(*connective_ptr) == typeid(expression::AND) ||
                                  typeid(*connective_ptr) == typeid(expression::OR))
                              {
                                  const auto left = Builder::extract_key_ranges(connective->left);
                                  const auto right = Builder::extract_key_ranges(connective->right);
                                  key_ranges.insert(left.begin(), left.end());
                                  key_ranges.insert(right.begin(), right.end());
                              }
                          },
                          [&key_ranges](const std::unique_ptr<expression::Atom> &atom) {
                              auto key = extract_key(atom);
                              if (key.has_value())
                              {
                                  auto atom_ptr = atom.get();
                                  if (typeid(*atom_ptr) == typeid(expression::EQ))
                                  {
                                      key_ranges.insert({key.value()});
                                  }
                                  else if (typeid(*atom_ptr) == typeid(expression::LT))
                                  {
                                      key_ranges.insert({std::numeric_limits<std::int64_t>::min(), key.value() - 1});
                                  }
                                  else if (typeid(*atom_ptr) == typeid(expression::LE))
                                  {
                                      key_ranges.insert({std::numeric_limits<std::int64_t>::min(), key.value()});
                                  }
                                  else if (typeid(*atom_ptr) == typeid(expression::GT))
                                  {
                                      key_ranges.insert({key.value() + 1, std::numeric_limits<std::int64_t>::max()});
                                  }
                                  else if (typeid(*atom_ptr) == typeid(expression::GE))
                                  {
                                      key_ranges.insert({key.value(), std::numeric_limits<std::int64_t>::max()});
                                  }
                              }
                          },
                          [](bool) {}},
               predicate);

    return key_ranges;
}

std::optional<std::int64_t> Builder::extract_key(const std::unique_ptr<beedb::expression::Atom> &atom)
{
    const auto &left = atom->left;
    const auto &right = atom->right;

    if (std::holds_alternative<std::int64_t>(left))
    {
        return {std::get<std::int64_t>(left)};
    }
    else if (std::holds_alternative<std::int64_t>(right))
    {
        return {std::get<std::int64_t>(right)};
    }

    return {};
}